# pylint: disable=no-name-in-module
from __future__ import annotations

import sys
from collections import defaultdict
from typing import DefaultDict, Literal

from bec_lib.logger import bec_logger
from qtpy.QtCore import QSize, Qt, QTimer
from qtpy.QtGui import QAction, QColor
from qtpy.QtWidgets import (
    QApplication,
    QComboBox,
    QLabel,
    QMainWindow,
    QMenu,
    QToolBar,
    QVBoxLayout,
    QWidget,
)

from bec_widgets.utils.colors import apply_theme, get_theme_name
from bec_widgets.utils.toolbars.actions import MaterialIconAction, ToolBarAction, WidgetAction
from bec_widgets.utils.toolbars.bundles import ToolbarBundle, ToolbarComponents
from bec_widgets.utils.toolbars.connections import BundleConnection

logger = bec_logger.logger

# Ensure that icons are shown in menus (especially on macOS)
QApplication.setAttribute(Qt.AA_DontShowIconsInMenus, False)


class ModularToolBar(QToolBar):
    """Modular toolbar with optional automatic initialization.

    Args:
        parent (QWidget, optional): The parent widget of the toolbar. Defaults to None.
        actions (dict, optional): A dictionary of action creators to populate the toolbar. Defaults to None.
        target_widget (QWidget, optional): The widget that the actions will target. Defaults to None.
        orientation (Literal["horizontal", "vertical"], optional): The initial orientation of the toolbar. Defaults to "horizontal".
        background_color (str, optional): The background color of the toolbar. Defaults to "rgba(0, 0, 0, 0)".
    """

    def __init__(
        self,
        parent=None,
        orientation: Literal["horizontal", "vertical"] = "horizontal",
        background_color: str = "rgba(0, 0, 0, 0)",
    ):
        super().__init__(parent=parent)

        self.background_color = background_color
        self.set_background_color(self.background_color)

        # Set the initial orientation
        self.set_orientation(orientation)

        self.components = ToolbarComponents(self)

        # Initialize bundles
        self.bundles: dict[str, ToolbarBundle] = {}
        self.shown_bundles: list[str] = []

        #########################
        # outdated items... remove
        self.available_widgets: DefaultDict[str, ToolBarAction] = defaultdict()
        ########################

    def new_bundle(self, name: str) -> ToolbarBundle:
        """
        Creates a new bundle component.

        Args:
            name (str): Unique identifier for the bundle.

        Returns:
            ToolbarBundle: The new bundle component.
        """
        if name in self.bundles:
            raise ValueError(f"Bundle with name '{name}' already exists.")
        bundle = ToolbarBundle(name=name, components=self.components)
        self.bundles[name] = bundle
        return bundle

    def add_bundle(self, bundle: ToolbarBundle):
        """
        Adds a bundle component to the toolbar.

        Args:
            bundle (ToolbarBundle): The bundle component to add.
        """
        if bundle.name in self.bundles:
            raise ValueError(f"Bundle with name '{bundle.name}' already exists.")
        self.bundles[bundle.name] = bundle
        if not bundle.bundle_actions:
            logger.warning(f"Bundle '{bundle.name}' has no actions.")

    def remove_bundle(self, name: str):
        """
        Removes a bundle component by name.

        Args:
            name (str): The name of the bundle to remove.
        """
        if name not in self.bundles:
            raise KeyError(f"Bundle with name '{name}' does not exist.")
        del self.bundles[name]
        if name in self.shown_bundles:
            self.shown_bundles.remove(name)
        logger.info(f"Bundle '{name}' removed from the toolbar.")

    def get_bundle(self, name: str) -> ToolbarBundle:
        """
        Retrieves a bundle component by name.

        Args:
            name (str): The name of the bundle to retrieve.

        Returns:
            ToolbarBundle: The bundle component.
        """
        if name not in self.bundles:
            raise KeyError(
                f"Bundle with name '{name}' does not exist. Available bundles: {list(self.bundles.keys())}"
            )
        return self.bundles[name]

    def show_bundles(self, bundle_names: list[str]):
        """
        Sets the bundles to be shown for the toolbar.

        Args:
            bundle_names (list[str]): A list of bundle names to show. If a bundle is not in this list, its actions will be hidden.
        """
        self.clear()
        for requested_bundle in bundle_names:
            bundle = self.get_bundle(requested_bundle)
            for bundle_action in bundle.bundle_actions.values():
                action = bundle_action()
                if action is None:
                    logger.warning(
                        f"Action for bundle '{requested_bundle}' has been deleted. Skipping."
                    )
                    continue
                action.add_to_toolbar(self, self.parent())
            separator = self.components.get_action_reference("separator")()
            if separator is not None:
                separator.add_to_toolbar(self, self.parent())
        self.update_separators()  # Ensure separators are updated after showing bundles
        self.shown_bundles = bundle_names

    def add_action(self, action_name: str, action: ToolBarAction):
        """
        Adds a single action to the toolbar. It will create a new bundle
        with the same name as the action.

        Args:
            action_name (str): Unique identifier for the action.
            action (ToolBarAction): The action to add.
        """
        self.components.add_safe(action_name, action)
        bundle = ToolbarBundle(name=action_name, components=self.components)
        bundle.add_action(action_name)
        self.add_bundle(bundle)

    def hide_action(self, action_name: str):
        """
        Hides a specific action in the toolbar.

        Args:
            action_name (str): Unique identifier for the action to hide.
        """
        action = self.components.get_action(action_name)
        if hasattr(action, "action") and action.action is not None:
            action.action.setVisible(False)
            self.update_separators()

    def show_action(self, action_name: str):
        """
        Shows a specific action in the toolbar.

        Args:
            action_name (str): Unique identifier for the action to show.
        """
        action = self.components.get_action(action_name)
        if hasattr(action, "action") and action.action is not None:
            action.action.setVisible(True)
            self.update_separators()

    @property
    def toolbar_actions(self) -> list[ToolBarAction]:
        """
        Returns a list of all actions currently in the toolbar.

        Returns:
            list[ToolBarAction]: List of actions in the toolbar.
        """
        actions = []
        for bundle in self.shown_bundles:
            if bundle not in self.bundles:
                continue
            for action in self.bundles[bundle].bundle_actions.values():
                action_instance = action()
                if action_instance is not None:
                    actions.append(action_instance)
        return actions

    def refresh(self):
        """Refreshes the toolbar by clearing and re-populating it."""
        self.clear()
        self.show_bundles(self.shown_bundles)

    def connect_bundle(self, connection_name: str, connector: BundleConnection):
        """
        Connects a bundle to a target widget or application.

        Args:
            bundle_name (str): The name of the bundle to connect.
            connector (BundleConnection): The connector instance that implements the connection logic.
        """
        bundle_name = connector.bundle_name
        if bundle_name not in self.bundles:
            raise KeyError(f"Bundle with name '{bundle_name}' does not exist.")
        connector.connect()
        self.bundles[bundle_name].add_connection(connection_name, connector)

    def disconnect_bundle(self, bundle_name: str, connection_name: str | None = None):
        """
        Disconnects a bundle connection.

        Args:
            bundle_name (str): The name of the bundle to disconnect.
            connection_name (str): The name of the connection to disconnect. If None, disconnects all connections for the bundle.
        """
        if bundle_name not in self.bundles:
            raise KeyError(f"Bundle with name '{bundle_name}' does not exist.")
        bundle = self.bundles[bundle_name]
        if connection_name is None:
            # Disconnect all connections in the bundle
            bundle.disconnect()
        else:
            bundle.remove_connection(name=connection_name)

    def set_background_color(self, color: str = "rgba(0, 0, 0, 0)"):
        """
        Sets the background color and other appearance settings.

        Args:
            color (str): The background color of the toolbar.
        """
        self.setIconSize(QSize(20, 20))
        self.setMovable(False)
        self.setFloatable(False)
        self.setContentsMargins(0, 0, 0, 0)
        self.background_color = color
        self.setStyleSheet(f"QToolBar {{ background-color: {color}; border: none; }}")

    def set_orientation(self, orientation: Literal["horizontal", "vertical"]):
        """Sets the orientation of the toolbar.

        Args:
            orientation (Literal["horizontal", "vertical"]): The desired orientation of the toolbar.
        """
        if orientation == "horizontal":
            self.setOrientation(Qt.Horizontal)
        elif orientation == "vertical":
            self.setOrientation(Qt.Vertical)
        else:
            raise ValueError("Orientation must be 'horizontal' or 'vertical'.")

    def update_material_icon_colors(self, new_color: str | tuple | QColor):
        """
        Updates the color of all MaterialIconAction icons.

        Args:
            new_color (str | tuple | QColor): The new color.
        """
        for action in self.available_widgets.values():
            if isinstance(action, MaterialIconAction):
                action.color = new_color
                updated_icon = action.get_icon()
                action.action.setIcon(updated_icon)

    def contextMenuEvent(self, event):
        """
        Overrides the context menu event to show toolbar actions with checkboxes and icons.

        Args:
            event (QContextMenuEvent): The context menu event.
        """
        menu = QMenu(self)
        theme = get_theme_name()
        if theme == "dark":
            menu.setStyleSheet("""
                QMenu {
                    background-color: rgba(50, 50, 50, 0.9);
                    border: 1px solid rgba(255, 255, 255, 0.2);
                }
                QMenu::item:selected {
                    background-color: rgba(0, 0, 255, 0.2);
                }
            """)
        else:
            # Light theme styling
            menu.setStyleSheet("""
                QMenu {
                    background-color: rgba(255, 255, 255, 0.9);
                    border: 1px solid rgba(0, 0, 0, 0.2);
                }
                QMenu::item:selected {
                    background-color: rgba(0, 0, 255, 0.2);
                }
            """)
        for ii, bundle in enumerate(self.shown_bundles):
            self.handle_bundle_context_menu(menu, bundle)
            if ii < len(self.shown_bundles) - 1:
                menu.addSeparator()
        menu.triggered.connect(self.handle_menu_triggered)
        menu.exec_(event.globalPos())

    def handle_bundle_context_menu(self, menu: QMenu, bundle_id: str):
        """
        Adds bundle actions to the context menu.

        Args:
            menu (QMenu): The context menu.
            bundle_id (str): The bundle identifier.
        """
        bundle = self.bundles.get(bundle_id)
        if not bundle:
            return
        for act_id in bundle.bundle_actions:
            toolbar_action = self.components.get_action(act_id)
            if not isinstance(toolbar_action, ToolBarAction) or not hasattr(
                toolbar_action, "action"
            ):
                continue
            qaction = toolbar_action.action
            if not isinstance(qaction, QAction):
                continue
            self._add_qaction_to_menu(menu, qaction, toolbar_action, act_id)

    def _add_qaction_to_menu(
        self, menu: QMenu, qaction: QAction, toolbar_action: ToolBarAction, act_id: str
    ):
        display_name = qaction.text() or toolbar_action.tooltip or act_id
        menu_action = QAction(display_name, self)
        menu_action.setCheckable(True)
        menu_action.setChecked(qaction.isVisible())
        menu_action.setData(act_id)  # Store the action_id

        # Set the icon if available
        if qaction.icon() and not qaction.icon().isNull():
            menu_action.setIcon(qaction.icon())
        menu.addAction(menu_action)

    def handle_action_context_menu(self, menu: QMenu, action_id: str):
        """
        Adds a single toolbar action to the context menu.

        Args:
            menu (QMenu): The context menu to which the action is added.
            action_id (str): Unique identifier for the action.
        """
        toolbar_action = self.available_widgets.get(action_id)
        if not isinstance(toolbar_action, ToolBarAction) or not hasattr(toolbar_action, "action"):
            return
        qaction = toolbar_action.action
        if not isinstance(qaction, QAction):
            return
        display_name = qaction.text() or toolbar_action.tooltip or action_id
        menu_action = QAction(display_name, self)
        menu_action.setCheckable(True)
        menu_action.setChecked(qaction.isVisible())
        menu_action.setIconVisibleInMenu(True)
        menu_action.setData(action_id)  # Store the action_id

        # Set the icon if available
        if qaction.icon() and not qaction.icon().isNull():
            menu_action.setIcon(qaction.icon())

        menu.addAction(menu_action)

    def handle_menu_triggered(self, action):
        """
        Handles the triggered signal from the context menu.

        Args:
            action: Action triggered.
        """
        action_id = action.data()
        if action_id:
            self.toggle_action_visibility(action_id)

    def toggle_action_visibility(self, action_id: str, visible: bool | None = None):
        """
        Toggles the visibility of a specific action.

        Args:
            action_id (str): Unique identifier.
            visible (bool): Whether the action should be visible. If None, toggles the current visibility.
        """
        if not self.components.exists(action_id):
            return
        tool_action = self.components.get_action(action_id)
        if hasattr(tool_action, "action") and tool_action.action is not None:
            if visible is None:
                visible = not tool_action.action.isVisible()
            tool_action.action.setVisible(visible)
        self.update_separators()

    def update_separators(self):
        """
        Hide separators that are adjacent to another separator, splitters, or have no non-separator actions between them.
        Splitters (ResizableSpacer) already provide visual separation, so we don't need separators next to them.
        """
        from bec_widgets.utils.toolbars.splitter import ResizableSpacer

        toolbar_actions = self.actions()

        # Helper function to check if a widget is a splitter
        def is_splitter_widget(action):
            widget = self.widgetForAction(action)
            return widget is not None and isinstance(widget, ResizableSpacer)

        # First pass: set visibility based on surrounding non-separator actions.
        for i, action in enumerate(toolbar_actions):
            if not action.isSeparator():
                continue
            prev_visible = None
            for j in range(i - 1, -1, -1):
                if toolbar_actions[j].isVisible():
                    prev_visible = toolbar_actions[j]
                    break
            next_visible = None
            for j in range(i + 1, len(toolbar_actions)):
                if toolbar_actions[j].isVisible():
                    next_visible = toolbar_actions[j]
                    break

            # Hide separator if adjacent to another separator, splitter, or at edges
            if (
                prev_visible is None
                or prev_visible.isSeparator()
                or is_splitter_widget(prev_visible)
            ) and (
                next_visible is None
                or next_visible.isSeparator()
                or is_splitter_widget(next_visible)
            ):
                action.setVisible(False)
            else:
                action.setVisible(True)
        # Second pass: ensure no two visible separators are adjacent, and no separators next to splitters.
        prev = None
        for action in toolbar_actions:
            if action.isVisible():
                if action.isSeparator():
                    # Hide separator if previous visible item was a separator or splitter
                    if prev and (prev.isSeparator() or is_splitter_widget(prev)):
                        action.setVisible(False)
                    else:
                        prev = action
                else:
                    prev = action

        if not toolbar_actions:
            return

        # Make sure the first visible action is not a separator
        for i, action in enumerate(toolbar_actions):
            if action.isVisible():
                if action.isSeparator():
                    action.setVisible(False)
                break

        # Make sure the last visible action is not a separator
        for i, action in enumerate(reversed(toolbar_actions)):
            if action.isVisible():
                if action.isSeparator():
                    action.setVisible(False)
                break

    def cleanup(self):
        """
        Cleans up the toolbar by removing all actions and bundles.
        """
        # First, disconnect all bundles
        for bundle_name in list(self.bundles.keys()):
            self.disconnect_bundle(bundle_name)

        # Clear all components
        self.components.cleanup()
        self.bundles.clear()


if __name__ == "__main__":  # pragma: no cover
    from bec_widgets.utils.toolbars.performance import PerformanceConnection, performance_bundle
    from bec_widgets.widgets.plots.toolbar_components.plot_export import plot_export_bundle

    class MainWindow(QMainWindow):  # pragma: no cover
        def __init__(self):
            super().__init__()
            self.setWindowTitle("Toolbar / ToolbarBundle Demo")
            self.central_widget = QWidget()
            self.setCentralWidget(self.central_widget)
            self.test_label = QLabel(text="Drag the splitter (⋮) to resize!")
            self.central_widget.layout = QVBoxLayout(self.central_widget)
            self.central_widget.layout.addWidget(self.test_label)

            self.toolbar = ModularToolBar(parent=self)
            self.addToolBar(self.toolbar)

            # Example: Bare combobox (no container). Give it a stable starting width
            self.example_combo = QComboBox(parent=self)
            self.example_combo.addItems(["device_1", "device_2", "device_3"])

            self.toolbar.components.add_safe(
                "example_combo", WidgetAction(widget=self.example_combo)
            )

            # Create a bundle with the combobox and a splitter
            self.bundle_combo_splitter = ToolbarBundle("example_combo", self.toolbar.components)
            self.bundle_combo_splitter.add_action("example_combo")
            # Add splitter; target the bare widget
            self.bundle_combo_splitter.add_splitter(
                name="splitter_example", target_widget=self.example_combo, min_width=100
            )

            # Add other bundles
            self.toolbar.add_bundle(self.bundle_combo_splitter)
            self.toolbar.add_bundle(performance_bundle(self.toolbar.components))
            self.toolbar.add_bundle(plot_export_bundle(self.toolbar.components))
            self.toolbar.connect_bundle(
                "base", PerformanceConnection(self.toolbar.components, self)
            )
            self.toolbar.components.add_safe(
                "text",
                MaterialIconAction(
                    "text_fields",
                    tooltip="Test Text Action",
                    checkable=True,
                    label_text="text",
                    text_position="under",
                ),
            )

            # Show bundles - notice how performance and plot_export appear compactly after splitter!
            self.toolbar.show_bundles(["example_combo", "performance", "plot_export"])
            self.toolbar.get_bundle("performance").add_action("save")
            self.toolbar.get_bundle("performance").add_action("text")
            self.toolbar.refresh()

            # Timer to disable and enable text button each 2s
            self.timer = QTimer()
            self.timer.timeout.connect(self.toggle_text_action)
            self.timer.start(2000)

        def toggle_text_action(self):
            text_action = self.toolbar.components.get_action("text")
            if text_action.action.isEnabled():
                text_action.action.setEnabled(False)
            else:
                text_action.action.setEnabled(True)

        def enable_fps_monitor(self, enabled: bool):
            """
            Example method to enable or disable FPS monitoring.
            This method should be implemented in the target widget.
            """
            if enabled:
                self.test_label.setText("FPS Monitor Enabled")
            else:
                self.test_label.setText("FPS Monitor Disabled")

    app = QApplication(sys.argv)
    apply_theme("light")
    main_window = MainWindow()
    main_window.show()
    sys.exit(app.exec_())



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Literal

from bec_lib.endpoints import EndpointInfo, MessageEndpoints
from bec_lib.logger import bec_logger
from pydantic import Field
from qtpy import QtCore, QtGui
from qtpy.QtGui import QColor
from qtpy.QtWidgets import QWidget

from bec_widgets import BECWidget
from bec_widgets.utils.bec_connector import ConnectionConfig
from bec_widgets.utils.colors import Colors
from bec_widgets.utils.error_popups import SafeProperty, SafeSlot

logger = bec_logger.logger
if TYPE_CHECKING:
    from bec_widgets.widgets.progress.ring_progress_bar.ring_progress_bar import (
        RingProgressContainerWidget,
    )


class ProgressbarConfig(ConnectionConfig):
    value: int | float = Field(0, description="Value for the progress bars.")
    direction: int = Field(
        -1, description="Direction of the progress bars. -1 for clockwise, 1 for counter-clockwise."
    )
    color: str | tuple = Field(
        (0, 159, 227, 255),
        description="Color for the progress bars. Can be tuple (R, G, B, A) or string HEX Code.",
    )
    background_color: str | tuple = Field(
        (200, 200, 200, 50),
        description="Background color for the progress bars. Can be tuple (R, G, B, A) or string HEX Code.",
    )
    link_colors: bool = Field(
        True,
        description="Whether to link the background color to the main color. If True, changing the main color will also change the background color.",
    )
    line_width: int = Field(20, description="Line widths for the progress bars.")
    start_position: int = Field(
        90,
        description="Start position for the progress bars in degrees. Default is 90 degrees - corresponds to "
        "the top of the ring.",
    )
    min_value: int | float = Field(0, description="Minimum value for the progress bars.")
    max_value: int | float = Field(100, description="Maximum value for the progress bars.")
    precision: int = Field(3, description="Precision for the progress bars.")
    mode: Literal["manual", "scan", "device"] = Field(
        "manual", description="Update mode for the progress bars."
    )
    device: str | None = Field(
        None,
        description="Device name for the device readback mode, only used when mode is 'device'.",
    )
    signal: str | None = Field(
        None,
        description="Signal name for the device readback mode, only used when mode is 'device'.",
    )


class Ring(BECWidget, QWidget):
    USER_ACCESS = [
        "set_value",
        "set_color",
        "set_background",
        "set_colors_linked",
        "set_line_width",
        "set_min_max_values",
        "set_start_angle",
        "set_update",
        "set_precision",
    ]
    RPC = True

    def __init__(self, parent: RingProgressContainerWidget | None = None, client=None, **kwargs):
        self.progress_container = parent
        self.config: ProgressbarConfig = ProgressbarConfig(widget_class=self.__class__.__name__)  # type: ignore
        super().__init__(parent=parent, client=client, config=self.config, **kwargs)
        self._color: QColor = self.convert_color(self.config.color)
        self._background_color: QColor = self.convert_color(self.config.background_color)
        self.registered_slot: tuple[Callable, str | EndpointInfo] | None = None
        self.RID = None
        self._gap = 5
        self._hovered = False
        self._hover_progress = 0.0
        self._hover_animation = QtCore.QPropertyAnimation(self, b"hover_progress", parent=self)
        self._hover_animation.setDuration(180)
        easing_curve = (
            QtCore.QEasingCurve.Type.OutCubic
            if hasattr(QtCore.QEasingCurve, "Type")
            else QtCore.QEasingCurve.Type.OutCubic
        )
        self._hover_animation.setEasingCurve(easing_curve)
        self.set_start_angle(self.config.start_position)

    def _request_update(self, *, refresh_tooltip: bool = True):
        # NOTE why not just overwrite update() to always refresh the tooltip?
        # Because in some cases (e.g. hover animation) we want to update the widget without refreshing the tooltip, to avoid performance issues.
        if refresh_tooltip:
            if self.progress_container and self.progress_container.is_ring_hovered(self):
                self.progress_container.refresh_hover_tooltip(self)
        self.update()

    def set_value(self, value: int | float):
        """
        Set the value for the ring widget

        Args:
            value(int | float): Value for the ring widget
        """
        self.value = value

    def set_color(self, color: str | tuple):
        """
        Set the color for the ring widget

        Args:
            color(str | tuple): Color for the ring widget. Can be HEX code or tuple (R, G, B, A).
        """
        self._color = self.convert_color(color)
        self.config.color = self._color.name()

        # Automatically set background color
        if self.config.link_colors:
            self._auto_set_background_color()

        self._request_update()

    def set_background(self, color: str | tuple | QColor):
        """
        Set the background color for the ring widget. The background color is only used when colors are not linked.

        Args:
            color(str | tuple): Background color for the ring widget. Can be HEX code or tuple (R, G, B, A).
        """
        # Only allow manual background color changes when colors are not linked
        if self.config.link_colors:
            return

        self._background_color = self.convert_color(color)
        self.config.background_color = self._background_color.name()
        self._request_update()

    def _auto_set_background_color(self):
        """
        Automatically set the background color based on the main color and the current theme.
        """
        palette = self.palette()
        bg = palette.color(QtGui.QPalette.ColorRole.Window)
        bg_color = Colors.subtle_background_color(self._color, bg)
        self.config.background_color = bg_color.name()
        self._background_color = bg_color
        self._request_update()

    def set_colors_linked(self, linked: bool):
        """
        Set whether the colors are linked for the ring widget.
        If colors are linked, changing the main color will also change the background color.

        Args:
            linked(bool): Whether to link the colors for the ring widget
        """
        self.config.link_colors = linked
        if linked:
            self._auto_set_background_color()
        self._request_update()

    def set_line_width(self, width: int):
        """
        Set the line width for the ring widget

        Args:
            width(int): Line width for the ring widget
        """
        self.config.line_width = width
        self._request_update()

    def set_min_max_values(self, min_value: int | float, max_value: int | float):
        """
        Set the min and max values for the ring widget.

        Args:
            min_value(int | float): Minimum value for the ring widget
            max_value(int | float): Maximum value for the ring widget
        """
        self.config.min_value = min_value
        self.config.max_value = max_value
        self._request_update()

    def set_start_angle(self, start_angle: int):
        """
        Set the start angle for the ring widget.

        Args:
            start_angle(int): Start angle for the ring widget in degrees
        """
        self.config.start_position = start_angle
        self._request_update()

    def set_update(
        self, mode: Literal["manual", "scan", "device"], device: str = "", signal: str = ""
    ):
        """
        Set the update mode for the ring widget.
        Modes:
        - "manual": Manual update mode, the value is set by the user.
        - "scan": Update mode for the scan progress. The value is updated by the current scan progress.
        - "device": Update mode for the device readback. The value is updated by the device readback. Take into account that user has to set the device name and limits.

        Args:
            mode(str): Update mode for the ring widget. Can be "manual", "scan" or "device"
            device(str): Device name for the device readback mode, only used when mode is "device"
            signal(str): Signal name for the device readback mode, only used when mode is "device"
        """

        match mode:
            case "manual":
                if self.config.mode == "manual":
                    return
                if self.registered_slot is not None:
                    self.bec_dispatcher.disconnect_slot(*self.registered_slot)
                self.config.mode = "manual"
                self.registered_slot = None
            case "scan":
                if self.config.mode == "scan":
                    return
                if self.registered_slot is not None:
                    self.bec_dispatcher.disconnect_slot(*self.registered_slot)
                self.config.mode = "scan"
                self.bec_dispatcher.connect_slot(
                    self.on_scan_progress, MessageEndpoints.scan_progress()
                )
                self.registered_slot = (self.on_scan_progress, MessageEndpoints.scan_progress())
            case "device":
                if self.registered_slot is not None:
                    self.bec_dispatcher.disconnect_slot(*self.registered_slot)
                self.config.mode = "device"
                if device == "":
                    self.registered_slot = None
                    return
                self.config.device = device
                # self.config.signal = self._get_signal_from_device(device, signal)
                signal = self._update_device_connection(device, signal)
                self.config.signal = signal

            case _:
                raise ValueError(f"Unsupported mode: {mode}")

    def set_precision(self, precision: int):
        """
        Set the precision for the ring widget.

        Args:
            precision(int): Precision for the ring widget
        """
        self.config.precision = precision
        self._request_update()

    def set_direction(self, direction: int):
        """
        Set the direction for the ring widget.

        Args:
            direction(int): Direction for the ring widget. -1 for clockwise, 1 for counter-clockwise.
        """
        self.config.direction = direction
        self._request_update()

    def _get_signals_for_device(self, device: str) -> dict[str, list[str]]:
        """
        Get the signals for the device.

        Args:
            device(str): Device name for the device

        Returns:
            dict[str, list[str]]: Dictionary with the signals for the device
        """
        dm = self.bec_dispatcher.client.device_manager
        if not dm:
            raise ValueError("Device manager is not available in the BEC client.")
        dev_obj = dm.devices.get(device)
        if dev_obj is None:
            raise ValueError(f"Device '{device}' not found in device manager.")

        progress_signals = [
            obj["component_name"]
            for obj in dev_obj._info["signals"].values()
            if obj["signal_class"] == "ProgressSignal"
        ]
        hinted_signals = [
            obj["obj_name"]
            for obj in dev_obj._info["signals"].values()
            if obj["kind_str"] == "hinted"
            and obj["signal_class"]
            not in ["ProgressSignal", "AsyncSignal", "AsyncMultiSignal", "DynamicSignal"]
        ]

        normal_signals = [
            obj["component_name"]
            for obj in dev_obj._info["signals"].values()
            if obj["kind_str"] == "normal"
        ]
        return {
            "progress_signals": progress_signals,
            "hinted_signals": hinted_signals,
            "normal_signals": normal_signals,
        }

    def _update_device_connection(self, device: str, signal: str | None) -> str:
        """
        Update the device connection for the ring widget.

        In general, we support two modes here:
        - If signal is provided, we use that directly.
        - If signal is not provided, we try to get the signal from the device manager.
          We first check for progress signals, then for hinted signals, and finally for normal signals.

        Depending on what type of signal we get (progress or hinted/normal), we subscribe to different endpoints.

        Args:
            device(str): Device name for the device mode
            signal(str): Signal name for the device mode

        Returns:
            str: The selected signal name for the device mode
        """
        logger.info(f"Updating device connection for device '{device}' and signal '{signal}'")
        dm = self.bec_dispatcher.client.device_manager
        if not dm:
            raise ValueError("Device manager is not available in the BEC client.")
        dev_obj = dm.devices.get(device)
        if dev_obj is None:
            return ""

        signals = self._get_signals_for_device(device)
        progress_signals = signals["progress_signals"]
        hinted_signals = signals["hinted_signals"]
        normal_signals = signals["normal_signals"]

        if not signal:
            # If signal is not provided, we try to get it from the device manager
            if len(progress_signals) > 0:
                signal = progress_signals[0]
                logger.info(
                    f"Using progress signal '{signal}' for device '{device}' in ring progress bar."
                )
            elif len(hinted_signals) > 0:
                signal = hinted_signals[0]
                logger.info(
                    f"Using hinted signal '{signal}' for device '{device}' in ring progress bar."
                )
            elif len(normal_signals) > 0:
                signal = normal_signals[0]
                logger.info(
                    f"Using normal signal '{signal}' for device '{device}' in ring progress bar."
                )
            else:
                logger.warning(f"No signals found for device '{device}' in ring progress bar.")
                return ""

        if signal in progress_signals:
            endpoint = MessageEndpoints.device_progress(device)
            self.bec_dispatcher.connect_slot(self.on_device_progress, endpoint)
            self.registered_slot = (self.on_device_progress, endpoint)
            return signal
        if signal in hinted_signals or signal in normal_signals:
            endpoint = MessageEndpoints.device_readback(device)
            self.bec_dispatcher.connect_slot(self.on_device_readback, endpoint)
            self.registered_slot = (self.on_device_readback, endpoint)
            return signal

    @SafeSlot(dict, dict)
    def on_scan_progress(self, msg, meta):
        """
        Update the ring widget with the scan progress.

        Args:
            msg(dict): Message with the scan progress
            meta(dict): Metadata for the message
        """
        current_RID = meta.get("RID", None)
        if current_RID != self.RID:
            self.set_min_max_values(0, msg.get("max_value", 100))
        self.set_value(msg.get("value", 0))
        self.update()

    @SafeSlot(dict, dict)
    def on_device_readback(self, msg, meta):
        """
        Update the ring widget with the device readback.

        Args:
            msg(dict): Message with the device readback
            meta(dict): Metadata for the message
        """
        device = self.config.device
        if device is None:
            return
        signal = self.config.signal or device
        value = msg.get("signals", {}).get(signal, {}).get("value", None)
        if value is None:
            return
        self.set_value(value)
        self.update()

    @SafeSlot(dict, dict)
    def on_device_progress(self, msg, meta):
        """
        Update the ring widget with the device progress.

        Args:
            msg(dict): Message with the device progress
            meta(dict): Metadata for the message
        """
        device = self.config.device
        if device is None:
            return
        max_val = msg.get("max_value", 100)
        self.set_min_max_values(0, max_val)
        value = msg.get("value", 0)
        if msg.get("done"):
            value = max_val
        self.set_value(value)
        self.update()

    def paintEvent(self, event):
        if not self.progress_container:
            return
        painter = QtGui.QPainter(self)
        painter.setRenderHint(QtGui.QPainter.RenderHint.Antialiasing)
        size = min(self.width(), self.height())

        # Center the ring
        x_offset = (self.width() - size) // 2
        y_offset = (self.height() - size) // 2

        max_ring_size = self.progress_container.get_max_ring_size()

        rect = QtCore.QRect(x_offset, y_offset, size, size)
        rect.adjust(max_ring_size, max_ring_size, -max_ring_size, -max_ring_size)

        # Background arc
        base_line_width = float(self.config.line_width)
        hover_line_delta = min(3.0, round(base_line_width * 0.6, 1))
        current_line_width = base_line_width + (hover_line_delta * self._hover_progress)
        painter.setPen(
            QtGui.QPen(self._background_color, current_line_width, QtCore.Qt.PenStyle.SolidLine)
        )

        gap: int = self.gap  # type: ignore

        # Important: Qt uses a 16th of a degree for angles. start_position is therefore multiplied by 16.
        start_position: float = self.config.start_position * 16  # type: ignore

        adjusted_rect = QtCore.QRectF(
            rect.left() + gap, rect.top() + gap, rect.width() - 2 * gap, rect.height() - 2 * gap
        )
        if self._hover_progress > 0.0:
            hover_radius_delta = 4.0
            base_radius = adjusted_rect.width() / 2
            if base_radius > 0:
                target_radius = base_radius + (hover_radius_delta * self._hover_progress)
                scale = target_radius / base_radius
                center = adjusted_rect.center()
                new_width = adjusted_rect.width() * scale
                new_height = adjusted_rect.height() * scale
                adjusted_rect = QtCore.QRectF(
                    center.x() - new_width / 2, center.y() - new_height / 2, new_width, new_height
                )
        painter.drawArc(adjusted_rect, start_position, 360 * 16)

        # Foreground arc
        pen = QtGui.QPen(self.color, current_line_width, QtCore.Qt.PenStyle.SolidLine)
        pen.setCapStyle(QtCore.Qt.PenCapStyle.RoundCap)
        painter.setPen(pen)
        proportion = (self.config.value - self.config.min_value) / (
            (self.config.max_value - self.config.min_value) + 1e-3
        )
        angle = int(proportion * 360 * 16 * self.config.direction)
        painter.drawArc(adjusted_rect, start_position, angle)
        painter.end()

    def set_hovered(self, hovered: bool):
        if hovered == self._hovered:
            return
        self._hovered = hovered
        self._hover_animation.stop()
        self._hover_animation.setStartValue(self._hover_progress)
        self._hover_animation.setEndValue(1.0 if hovered else 0.0)
        self._hover_animation.start()

    @staticmethod
    def convert_color(color: str | tuple | QColor) -> QColor:
        """
        Convert the color to QColor

        Args:
            color(str | tuple | QColor): Color for the ring widget. Can be HEX code or tuple (R, G, B, A) or QColor.
        """

        if isinstance(color, QColor):
            return color
        if isinstance(color, str):
            return QtGui.QColor(color)
        if isinstance(color, (tuple, list)):
            return QtGui.QColor(*color)
        raise ValueError(f"Unsupported color format: {color}")

    def cleanup(self):
        """
        Cleanup the ring widget.
        Disconnect any registered slots.
        """
        if self.registered_slot is not None:
            self.bec_dispatcher.disconnect_slot(*self.registered_slot)
            self.registered_slot = None

    ###############################################
    ####### QProperties ###########################
    ###############################################

    @SafeProperty(int)
    def gap(self) -> int:
        return self._gap

    @gap.setter
    def gap(self, value: int):
        self._gap = value
        self._request_update()

    @SafeProperty(bool)
    def link_colors(self) -> bool:
        return self.config.link_colors

    @link_colors.setter
    def link_colors(self, value: bool):
        logger.info(f"Setting link_colors to {value}")
        self.set_colors_linked(value)

    @SafeProperty(QColor)
    def color(self) -> QColor:
        return self._color

    @color.setter
    def color(self, value: QColor):
        self.set_color(value)

    @SafeProperty(QColor)
    def background_color(self) -> QColor:
        return self._background_color

    @background_color.setter
    def background_color(self, value: QColor):
        self.set_background(value)

    @SafeProperty(float)
    def value(self) -> float:
        return self.config.value

    @value.setter
    def value(self, value: float):
        self.config.value = round(
            float(max(self.config.min_value, min(self.config.max_value, value))),
            self.config.precision,
        )
        self._request_update()

    @SafeProperty(float)
    def min_value(self) -> float:
        return self.config.min_value

    @min_value.setter
    def min_value(self, value: float):
        self.config.min_value = value
        self._request_update()

    @SafeProperty(float)
    def max_value(self) -> float:
        return self.config.max_value

    @max_value.setter
    def max_value(self, value: float):
        self.config.max_value = value
        self._request_update()

    @SafeProperty(str)
    def mode(self) -> str:
        return self.config.mode

    @mode.setter
    def mode(self, value: str):
        self.set_update(value)
        self._request_update()

    @SafeProperty(str)
    def device(self) -> str:
        return self.config.device or ""

    @device.setter
    def device(self, value: str):
        self.config.device = value
        self._request_update()

    @SafeProperty(str)
    def signal(self) -> str:
        return self.config.signal or ""

    @signal.setter
    def signal(self, value: str):
        self.config.signal = value
        self._request_update()

    @SafeProperty(int)
    def line_width(self) -> int:
        return self.config.line_width

    @line_width.setter
    def line_width(self, value: int):
        self.config.line_width = value
        self._request_update()

    @SafeProperty(int)
    def start_position(self) -> int:
        return self.config.start_position

    @start_position.setter
    def start_position(self, value: int):
        self.config.start_position = value
        self._request_update()

    @SafeProperty(int)
    def precision(self) -> int:
        return self.config.precision

    @precision.setter
    def precision(self, value: int):
        self.config.precision = value
        self._request_update()

    @SafeProperty(int)
    def direction(self) -> int:
        return self.config.direction

    @direction.setter
    def direction(self, value: int):
        self.config.direction = value
        self._request_update()

    @SafeProperty(float)
    def hover_progress(self) -> float:
        return self._hover_progress

    @hover_progress.setter
    def hover_progress(self, value: float):
        self._hover_progress = value
        self._request_update(refresh_tooltip=False)

    def cleanup(self):
        """
        Cleanup the ring widget.
        Disconnect any registered slots.
        """
        if self.registered_slot is not None:
            self.bec_dispatcher.disconnect_slot(*self.registered_slot)
            self.registered_slot = None
        self._hover_animation.stop()
        super().cleanup()


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtpy.QtWidgets import QApplication

    from bec_widgets.utils.colors import apply_theme

    app = QApplication(sys.argv)
    apply_theme("dark")
    ring = Ring()
    ring.export_settings()
    ring.show()
    sys.exit(app.exec())



import sys
from typing import Literal, Optional

from qtpy.QtCore import Property, QEasingCurve, QPropertyAnimation
from qtpy.QtGui import QAction
from qtpy.QtWidgets import (
    QApplication,
    QFrame,
    QHBoxLayout,
    QLabel,
    QMainWindow,
    QScrollArea,
    QSizePolicy,
    QStackedWidget,
    QVBoxLayout,
    QWidget,
)

from bec_widgets.utils.toolbars.bundles import ToolbarBundle
from bec_widgets.utils.toolbars.toolbar import MaterialIconAction, ModularToolBar


class SidePanel(QWidget):
    """
    Side panel widget that can be placed on the left, right, top, or bottom of the main widget.
    """

    def __init__(
        self,
        parent=None,
        orientation: Literal["left", "right", "top", "bottom"] = "left",
        panel_max_width: int = 200,
        animation_duration: int = 200,
        animations_enabled: bool = True,
        show_toolbar: bool = True,
    ):
        super().__init__(parent=parent)

        self.setProperty("skip_settings", True)

        self._orientation = orientation
        self._panel_max_width = panel_max_width
        self._animation_duration = animation_duration
        self._animations_enabled = animations_enabled
        self._show_toolbar = show_toolbar

        self._panel_width = 0
        self._panel_height = 0
        self.panel_visible = False
        self.current_action: Optional[QAction] = None
        self.current_index: Optional[int] = None
        self.switching_actions = False

        self._init_ui()

    def _init_ui(self):
        """
        Initialize the UI elements.
        """
        if self._orientation in ("left", "right"):
            self.main_layout = QHBoxLayout(self)
            self.main_layout.setContentsMargins(0, 0, 0, 0)
            self.main_layout.setSpacing(0)

            self.toolbar = ModularToolBar(parent=self, orientation="vertical")

            self.container = QWidget()
            self.container.layout = QVBoxLayout(self.container)
            self.container.layout.setContentsMargins(0, 0, 0, 0)
            self.container.layout.setSpacing(0)

            self.stack_widget = QStackedWidget()
            self.stack_widget.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Expanding)
            self.stack_widget.setMinimumWidth(5)
            self.stack_widget.setMaximumWidth(self._panel_max_width)

            if self._orientation in ("left", "right"):
                if self._show_toolbar:
                    self.main_layout.addWidget(self.toolbar)

                if self._orientation == "left":
                    self.main_layout.addWidget(self.container)
                else:
                    self.main_layout.insertWidget(0, self.container)
            self.container.layout.addWidget(self.stack_widget)

            self.menu_anim = QPropertyAnimation(self, b"panel_width")
            self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Expanding)
            self.panel_width = 0  # start hidden

        else:
            self.main_layout = QVBoxLayout(self)
            self.main_layout.setContentsMargins(0, 0, 0, 0)
            self.main_layout.setSpacing(0)

            self.toolbar = ModularToolBar(parent=self, orientation="horizontal")

            self.container = QWidget()
            self.container.layout = QVBoxLayout(self.container)
            self.container.layout.setContentsMargins(0, 0, 0, 0)
            self.container.layout.setSpacing(0)

            self.stack_widget = QStackedWidget()
            self.stack_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
            self.stack_widget.setMinimumHeight(5)
            self.stack_widget.setMaximumHeight(self._panel_max_width)

            if self._orientation == "top":
                if self._show_toolbar:
                    self.main_layout.addWidget(self.toolbar)
                self.main_layout.addWidget(self.container)
            else:
                self.main_layout.addWidget(self.container)
                if self._show_toolbar:
                    self.main_layout.addWidget(self.toolbar)

            self.container.layout.addWidget(self.stack_widget)

            self.menu_anim = QPropertyAnimation(self, b"panel_height")
            self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
            self.panel_height = 0  # start hidden

        self.menu_anim.setDuration(self._animation_duration)
        self.menu_anim.setEasingCurve(QEasingCurve.InOutQuad)

    @Property(int)
    def panel_width(self):
        """Get the panel width."""
        return self._panel_width

    @panel_width.setter
    def panel_width(self, width: int):
        """Set the panel width."""
        self._panel_width = width
        if self._orientation in ("left", "right"):
            self.stack_widget.setFixedWidth(width)

    @Property(int)
    def panel_height(self):
        """Get the panel height."""
        return self._panel_height

    @panel_height.setter
    def panel_height(self, height: int):
        """Set the panel height."""
        self._panel_height = height
        if self._orientation in ("top", "bottom"):
            self.stack_widget.setFixedHeight(height)

    @Property(int)
    def panel_max_width(self):
        """Get the maximum width of the panel."""
        return self._panel_max_width

    @panel_max_width.setter
    def panel_max_width(self, size: int):
        """Set the maximum width of the panel."""
        self._panel_max_width = size
        if self._orientation in ("left", "right"):
            self.stack_widget.setMaximumWidth(self._panel_max_width)
        else:
            self.stack_widget.setMaximumHeight(self._panel_max_width)

    @Property(int)
    def animation_duration(self):
        """Get the duration of the animation."""
        return self._animation_duration

    @animation_duration.setter
    def animation_duration(self, duration: int):
        """Set the duration of the animation."""
        self._animation_duration = duration
        self.menu_anim.setDuration(duration)

    @Property(bool)
    def animations_enabled(self):
        """Get the status of the animations."""
        return self._animations_enabled

    @animations_enabled.setter
    def animations_enabled(self, enabled: bool):
        """Set the status of the animations."""
        self._animations_enabled = enabled

    def show_panel(self, idx: int):
        """
        Show the side panel with animation and switch to idx.
        """
        self.stack_widget.setCurrentIndex(idx)
        self.panel_visible = True
        self.current_index = idx

        if self._orientation in ("left", "right"):
            start_val, end_val = 0, self._panel_max_width
        else:
            start_val, end_val = 0, self._panel_max_width

        if self._animations_enabled:
            self.menu_anim.stop()
            self.menu_anim.setStartValue(start_val)
            self.menu_anim.setEndValue(end_val)
            self.menu_anim.start()
        else:
            if self._orientation in ("left", "right"):
                self.panel_width = end_val
            else:
                self.panel_height = end_val

    def hide_panel(self):
        """
        Hide the side panel with animation.
        """
        self.panel_visible = False
        self.current_index = None

        if self._orientation in ("left", "right"):
            start_val, end_val = self._panel_max_width, 0
        else:
            start_val, end_val = self._panel_max_width, 0

        if self._animations_enabled:
            self.menu_anim.stop()
            self.menu_anim.setStartValue(start_val)
            self.menu_anim.setEndValue(end_val)
            self.menu_anim.start()
        else:
            if self._orientation in ("left", "right"):
                self.panel_width = end_val
            else:
                self.panel_height = end_val

    def switch_to(self, idx: int):
        """
        Switch to the specified index without animation.
        """
        if self.current_index != idx:
            self.stack_widget.setCurrentIndex(idx)
            self.current_index = idx

    def add_menu(
        self,
        widget: QWidget,
        action_id: str | None = None,
        icon_name: str | None = None,
        tooltip: str | None = None,
        title: str | None = None,
    ) -> int:
        """
        Add a menu to the side panel.

        Args:
            widget(QWidget): The widget to add to the panel.
            action_id(str | None): The ID of the action. Optional if no toolbar action is needed.
            icon_name(str | None): The name of the icon. Optional if no toolbar action is needed.
            tooltip(str | None): The tooltip for the action. Optional if no toolbar action is needed.
            title(str | None): The title of the panel.

        Returns:
            int: The index of the added panel, which can be used with show_panel() and switch_to().
        """
        # container_widget: top-level container for the stacked page
        container_widget = QWidget()
        container_layout = QVBoxLayout(container_widget)
        container_layout.setContentsMargins(0, 0, 0, 0)
        container_layout.setSpacing(5)

        if title is not None:
            title_label = QLabel(f"<b>{title}</b>")
            title_label.setStyleSheet("font-size: 16px;")
            container_layout.addWidget(title_label)

        # Create a QScrollArea for the actual widget to ensure scrolling if the widget inside is too large
        scroll_area = QScrollArea()
        scroll_area.setFrameShape(QFrame.NoFrame)
        scroll_area.setWidgetResizable(True)
        # Let the scroll area expand in both directions if there's room
        scroll_area.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        scroll_area.setWidget(widget)

        # Put the scroll area in the container layout
        container_layout.addWidget(scroll_area)

        # Optionally stretch the scroll area to fill vertical space
        container_layout.setStretchFactor(scroll_area, 1)

        # Add container_widget to the stacked widget
        index = self.stack_widget.count()
        self.stack_widget.addWidget(container_widget)

        # Add an action to the toolbar if action_id, icon_name, and tooltip are provided
        if action_id is not None and icon_name is not None and tooltip is not None:
            action = MaterialIconAction(
                icon_name=icon_name, tooltip=tooltip, checkable=True, parent=self
            )
            self.toolbar.components.add_safe(action_id, action)
            bundle = ToolbarBundle(action_id, self.toolbar.components)
            bundle.add_action(action_id)
            self.toolbar.add_bundle(bundle)
            shown_bundles = self.toolbar.shown_bundles
            shown_bundles.append(action_id)
            self.toolbar.show_bundles(shown_bundles)

            def on_action_toggled(checked: bool):
                if self.switching_actions:
                    return

                if checked:
                    if self.current_action and self.current_action != action.action:
                        self.switching_actions = True
                        self.current_action.setChecked(False)
                        self.switching_actions = False

                    self.current_action = action.action

                    if not self.panel_visible:
                        self.show_panel(index)
                    else:
                        self.switch_to(index)
                else:
                    if self.current_action == action.action:
                        self.current_action = None
                        self.hide_panel()

            action.action.toggled.connect(on_action_toggled)

        return index


############################################
# DEMO APPLICATION
############################################


class ExampleApp(QMainWindow):  # pragma: no cover
    def __init__(self):
        super().__init__()
        self.setWindowTitle("Side Panel Example")

        central_widget = QWidget()
        self.setCentralWidget(central_widget)
        self.layout = QHBoxLayout(central_widget)

        # Create side panel
        self.side_panel = SidePanel(self, orientation="left", panel_max_width=250)
        self.layout.addWidget(self.side_panel)

        from bec_widgets.widgets.plots.waveform.waveform import Waveform

        self.plot = Waveform()
        self.layout.addWidget(self.plot)

        self.add_side_menus()

    def add_side_menus(self):
        # Example 1: With action, icon, and tooltip
        widget1 = QWidget()
        layout1 = QVBoxLayout(widget1)
        for i in range(15):
            layout1.addWidget(QLabel(f"Widget 1 label row {i}"))
        self.side_panel.add_menu(
            widget=widget1,
            action_id="widget1",
            icon_name="counter_1",
            tooltip="Show Widget 1",
            title="Widget 1 Panel",
        )

        # Example 2: With action, icon, and tooltip
        widget2 = QWidget()
        layout2 = QVBoxLayout(widget2)
        layout2.addWidget(QLabel("Short widget 2 content"))
        self.side_panel.add_menu(
            widget=widget2,
            action_id="widget2",
            icon_name="counter_2",
            tooltip="Show Widget 2",
            title="Widget 2 Panel",
        )

        # Example 3: With action, icon, and tooltip
        widget3 = QWidget()
        layout3 = QVBoxLayout(widget3)
        for i in range(10):
            layout3.addWidget(QLabel(f"Line {i} for Widget 3"))
        self.side_panel.add_menu(
            widget=widget3,
            action_id="widget3",
            icon_name="counter_3",
            tooltip="Show Widget 3",
            title="Widget 3 Panel",
        )

        # Example 4: Without action, icon, and tooltip (can only be shown programmatically)
        widget4 = QWidget()
        layout4 = QVBoxLayout(widget4)
        layout4.addWidget(QLabel("This panel has no toolbar button"))
        layout4.addWidget(QLabel("It can only be shown programmatically"))
        self.hidden_panel_index = self.side_panel.add_menu(widget=widget4, title="Hidden Panel")

        # Example of how to show the hidden panel programmatically after 3 seconds
        from qtpy.QtCore import QTimer

        QTimer.singleShot(3000, lambda: self.side_panel.show_panel(self.hidden_panel_index))


if __name__ == "__main__":  # pragma: no cover
    app = QApplication(sys.argv)
    window = ExampleApp()
    window.resize(1000, 700)
    window.show()
    sys.exit(app.exec())



import os
import sys

import bec_widgets.widgets.containers.qt_ads as QtAds
from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.error_popups import SafeProperty, SafeSlot

if sys.platform.startswith("linux"):
    qt_platform = os.environ.get("QT_QPA_PLATFORM", "")
    if qt_platform != "offscreen":
        os.environ["QT_QPA_PLATFORM"] = "xcb"

# Default QtAds configuration
QtAds.CDockManager.setConfigFlag(QtAds.CDockManager.eConfigFlag.FocusHighlighting, True)
QtAds.CDockManager.setConfigFlag(
    QtAds.CDockManager.eConfigFlag.RetainTabSizeWhenCloseButtonHidden, True
)

__all__ = ["BECWidget", "SafeSlot", "SafeProperty"]



from qtpy.QtWidgets import QHBoxLayout, QSizePolicy, QWidget

from bec_widgets.utils.toolbars.actions import WidgetAction
from bec_widgets.utils.toolbars.bundles import ToolbarBundle, ToolbarComponents
from bec_widgets.utils.toolbars.connections import BundleConnection
from bec_widgets.widgets.control.device_input.device_combobox.device_combobox import DeviceComboBox
from bec_widgets.widgets.control.device_input.signal_combobox.signal_combobox import SignalComboBox


class DeviceSelection(QWidget):
    """Device and signal selection widget for image toolbar."""

    def __init__(self, parent=None, client=None):
        super().__init__(parent=parent)

        self.client = client
        self.supported_signals = [
            "PreviewSignal",
            "AsyncSignal",
            "AsyncMultiSignal",
            "DynamicSignal",
        ]

        # Create device combobox with signal class filter
        # This will only show devices that have signals matching the supported signal classes
        self.device_combo_box = DeviceComboBox(
            parent=self, client=self.client, signal_class_filter=self.supported_signals
        )
        self.device_combo_box.setToolTip("Select Device")
        self.device_combo_box.setEditable(True)
        # Set expanding size policy so it grows with available space
        self.device_combo_box.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        self.device_combo_box.lineEdit().setPlaceholderText("Select Device")

        # Configure SignalComboBox to filter by PreviewSignal and supported async signals
        # Also filter by ndim (1D and 2D only) for Image widget
        self.signal_combo_box = SignalComboBox(
            parent=self,
            client=self.client,
            signal_class_filter=[
                "PreviewSignal",
                "AsyncSignal",
                "AsyncMultiSignal",
                "DynamicSignal",
            ],
            ndim_filter=[1, 2],  # Only show 1D and 2D signals for Image widget
            store_signal_config=True,
            require_device=True,
        )
        self.signal_combo_box.setToolTip("Select Signal")
        self.signal_combo_box.setEditable(True)
        # Set expanding size policy so it grows with available space
        self.signal_combo_box.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        self.signal_combo_box.lineEdit().setPlaceholderText("Select Signal")

        # Connect comboboxes together
        self.device_combo_box.currentTextChanged.connect(self.signal_combo_box.set_device)
        self.device_combo_box.device_reset.connect(self.signal_combo_box.reset_selection)

        # Simple horizontal layout with stretch to fill space
        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(2)
        layout.addWidget(self.device_combo_box, stretch=1)
        layout.addWidget(self.signal_combo_box, stretch=1)

    def set_device_and_signal(self, device: str | None, signal: str | None) -> None:
        """Set the displayed device and signal without emitting selection signals."""
        device = device or ""
        signal = signal or ""

        self.device_combo_box.blockSignals(True)
        self.signal_combo_box.blockSignals(True)

        try:
            if device:
                # Set device in device_combo_box
                index = self.device_combo_box.findText(device)
                if index >= 0:
                    self.device_combo_box.setCurrentIndex(index)
                else:
                    # Device not found in list, but still set it
                    self.device_combo_box.setCurrentText(device)

                # Only update signal combobox device filter if it's actually changing
                # This prevents redundant repopulation which can cause duplicates !!!!
                current_device = getattr(self.signal_combo_box, "_device", None)
                if current_device != device:
                    self.signal_combo_box.set_device(device)

                # Sync signal combobox selection
                if signal:
                    # Try to find the signal by component_name (which is what's displayed)
                    found = False
                    for i in range(self.signal_combo_box.count()):
                        text = self.signal_combo_box.itemText(i)
                        config_data = self.signal_combo_box.itemData(i)

                        # Check if this matches our signal
                        if config_data:
                            component_name = config_data.get("component_name", "")
                            if text == component_name or text == signal:
                                self.signal_combo_box.setCurrentIndex(i)
                                found = True
                                break

                    if not found:
                        # Fallback: try to match the signal directly
                        index = self.signal_combo_box.findText(signal)
                        if index >= 0:
                            self.signal_combo_box.setCurrentIndex(index)
            else:
                # No device set, clear selections
                self.device_combo_box.setCurrentText("")
                self.signal_combo_box.reset_selection()
        finally:
            # Always unblock signals
            self.device_combo_box.blockSignals(False)
            self.signal_combo_box.blockSignals(False)

    def set_connection_status(self, status: str, message: str | None = None) -> None:
        tooltip = f"Connection status: {status}"
        if message:
            tooltip = f"{tooltip}\n{message}"
        self.device_combo_box.setToolTip(tooltip)
        self.signal_combo_box.setToolTip(tooltip)

        if not self.device_combo_box.is_valid_input or not self.signal_combo_box.is_valid_input:
            return

        if status == "error":
            style = "border: 1px solid orange;"
        else:
            style = "border: 1px solid transparent;"

        self.device_combo_box.setStyleSheet(style)
        self.signal_combo_box.setStyleSheet(style)

    def cleanup(self):
        """Clean up the widget resources."""
        self.device_combo_box.close()
        self.device_combo_box.deleteLater()
        self.signal_combo_box.close()
        self.signal_combo_box.deleteLater()


def device_selection_bundle(components: ToolbarComponents, client=None) -> ToolbarBundle:
    """
    Creates a device selection toolbar bundle for Image widget.

    Includes a resizable splitter after the device selection. All subsequent bundles'
    actions will appear compactly after the splitter with no gaps.

    Args:
        components (ToolbarComponents): The components to be added to the bundle.
        client: The BEC client instance.

    Returns:
        ToolbarBundle: The device selection toolbar bundle.
    """
    device_selection_widget = DeviceSelection(parent=components.toolbar, client=client)
    components.add_safe(
        "device_selection", WidgetAction(widget=device_selection_widget, adjust_size=False)
    )

    bundle = ToolbarBundle("device_selection", components)
    bundle.add_action("device_selection")

    bundle.add_splitter(
        name="device_selection_splitter",
        target_widget=device_selection_widget,
        min_width=210,
        max_width=600,
    )

    return bundle


class DeviceSelectionConnection(BundleConnection):
    """
    Connection helper for the device selection bundle.
    """

    def __init__(self, components: ToolbarComponents, target_widget=None):
        super().__init__(parent=components.toolbar)
        self.bundle_name = "device_selection"
        self.components = components
        self.target_widget = target_widget
        self._connected = False
        self.register_property_sync("device", self._sync_from_device)
        self.register_property_sync("signal", self._sync_from_signal)
        self.register_property_sync("connection_status", self._sync_connection_status)
        self.register_property_sync("connection_error", self._sync_connection_status)

    def _widget(self) -> DeviceSelection:
        return self.components.get_action("device_selection").widget

    def connect(self):
        if self._connected:
            return
        widget = self._widget()
        widget.device_combo_box.device_selected.connect(
            self.target_widget.on_device_selection_changed
        )
        widget.signal_combo_box.device_signal_changed.connect(
            self.target_widget.on_device_selection_changed
        )
        self.connect_property_sync(self.target_widget)
        self._connected = True

    def disconnect(self):
        if not self._connected:
            return
        widget = self._widget()
        widget.device_combo_box.device_selected.disconnect(
            self.target_widget.on_device_selection_changed
        )
        widget.signal_combo_box.device_signal_changed.disconnect(
            self.target_widget.on_device_selection_changed
        )
        self.disconnect_property_sync(self.target_widget)
        self._connected = False
        widget.cleanup()

    def _sync_from_device(self, _):
        try:
            widget = self._widget()
        except Exception:
            return

        widget.set_device_and_signal(self.target_widget.device, self.target_widget.signal)
        self.target_widget._sync_signal_from_toolbar()

    def _sync_from_signal(self, _):
        try:
            widget = self._widget()
        except Exception:
            return

        widget.set_device_and_signal(self.target_widget.device, self.target_widget.signal)

    def _sync_connection_status(self, _):
        try:
            widget = self._widget()
        except Exception:
            return

        widget.set_connection_status(
            self.target_widget._config.connection_status,
            self.target_widget._config.connection_error,
        )



from bec_qthemes import material_icon
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QHBoxLayout, QPushButton, QSizePolicy, QToolButton, QWidget

from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.error_popups import SafeSlot


class StopButton(BECWidget, QWidget):
    """A button that stops the current scan."""

    PLUGIN = True
    ICON_NAME = "dangerous"
    RPC = False

    def __init__(self, parent=None, client=None, config=None, gui_id=None, toolbar=False, **kwargs):
        super().__init__(parent=parent, client=client, gui_id=gui_id, config=config, **kwargs)

        self.get_bec_shortcuts()

        self.layout = QHBoxLayout(self)
        self.layout.setSpacing(0)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setAlignment(Qt.AlignmentFlag.AlignVCenter)

        if toolbar:
            icon = material_icon("stop", color="#cc181e", filled=True, convert_to_pixmap=False)
            self.button = QToolButton(icon=icon)
            self.button.setToolTip("Stop the scan queue")
        else:
            self.button = QPushButton()
            self.button.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
            self.button.setText("Stop")
            self.button.setProperty("variant", "danger")
        self.button.clicked.connect(self.stop_scan)

        self.layout.addWidget(self.button)

    @SafeSlot()
    def stop_scan(
        self,
    ):  # , scan_id: str | None = None): #FIXME scan_id will be added when combining with Queue widget
        """
        Stop the scan.

        Args:
            scan_id(str|None): The scan id to stop. If None, the current scan will be stopped.
        """
        self.queue.request_scan_halt()


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtpy.QtWidgets import QApplication, QVBoxLayout, QWidget

    from bec_widgets.widgets.control.buttons.stop_button.stop_button import StopButton

    class MyGui(QWidget):
        def __init__(self):
            super().__init__()
            self.setLayout(QVBoxLayout())
            # Create and add the StopButton to the layout
            self.stop_button = StopButton()
            self.layout().addWidget(self.stop_button)

    # Example of how this custom GUI might be used:
    app = QApplication([])
    my_gui = MyGui()
    my_gui.show()
    sys.exit(app.exec_())



"""Experiment Selection View for BEC Atlas Admin Widget"""

from datetime import datetime
from typing import Any

from bec_lib.logger import bec_logger
from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import (
    QCheckBox,
    QHBoxLayout,
    QHeaderView,
    QLabel,
    QLineEdit,
    QSizePolicy,
    QTableWidget,
    QTableWidgetItem,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)
from thefuzz import fuzz

from bec_widgets.utils.error_popups import SafeSlot
from bec_widgets.utils.fuzzy_search import is_match
from bec_widgets.widgets.services.bec_atlas_admin_view.experiment_selection.experiment_mat_card import (
    ExperimentMatCard,
)
from bec_widgets.widgets.services.bec_atlas_admin_view.experiment_selection.utils import (
    format_name,
    format_schedule,
)

logger = bec_logger.logger


class ExperimentSelection(QWidget):
    experiment_selected = Signal(dict)

    def __init__(self, experiment_infos=None, parent=None):
        super().__init__(parent=parent)
        self._experiment_infos = experiment_infos or []
        self._next_experiment = self._select_next_experiment(self._experiment_infos)
        self._enable_fuzzy_search: bool = True
        self._hidden_rows: set[int] = set()
        self._headers: dict[str, str] = {
            "pgroup": "pgroup",
            "title": "Title",
            "name": "Name",
            "schedule_start": "Schedule (start)",
            "schedule_end": "Schedule (end)",
        }
        self._table_infos: list[dict[str, Any]] = []

        main_layout = QVBoxLayout(self)
        main_layout.setContentsMargins(0, 0, 0, 0)
        # main_layout.setSpacing(12)
        main_layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self.setAutoFillBackground(True)

        self._tabs = QTabWidget(self)
        main_layout.addWidget(self._tabs, stretch=1)

        self._card_tab = ExperimentMatCard(
            parent=self, show_activate_button=True, button_text="Activate"
        )
        self._card_tab.experiment_selected.connect(self._emit_selected_experiment)
        if self._next_experiment:
            self._card_tab.set_experiment_info(self._next_experiment)
        self._table_tab = QWidget(self)
        self._tabs.addTab(self._card_tab, "Next Experiment")
        self._tabs.addTab(self._table_tab, "Manual Selection")

        self._build_table_tab()
        self._tabs.currentChanged.connect(self._on_tab_changed)

        button_layout = QHBoxLayout()
        main_layout.addLayout(button_layout)
        self._apply_table_filters()
        self.restore_default_view()

    def restore_default_view(self):
        """Reset the view to the default state, showing the next experiment card."""
        self._tabs.setCurrentWidget(self._card_tab)

    def set_experiment_infos(self, experiment_infos: list[dict]):
        """
        Update the experiment information displayed in the view. It will in addition determine
        the next experiment to be shown in the card view. If no next experiment can be determined,
        the card view will be cleared.

        Args:
            experiment_infos (list[dict]): A list of experiment information dictionaries.
        """
        self._experiment_infos = experiment_infos
        self._next_experiment = self._select_next_experiment(self._experiment_infos)
        if self._next_experiment:
            self._card_tab.set_experiment_info(self._next_experiment)
        else:
            self._card_tab.clear_experiment_info()
        self._apply_table_filters()

    def _setup_search(self, layout: QVBoxLayout):
        """
        Create components related to the search functionality

        Args:
            layout (QVBoxLayout): The layout to which the search components will be added.
        """

        # Create search bar
        search_layout = QHBoxLayout()
        self.search_label = QLabel("Search:")
        self.search_input = QLineEdit()
        self.search_input.setPlaceholderText("Filter experiments...")
        self.search_input.setClearButtonEnabled(True)
        self.search_input.textChanged.connect(self._apply_row_filter)
        search_layout.addWidget(self.search_label)
        search_layout.addWidget(self.search_input)

        # Add exact match toggle
        fuzzy_layout = QHBoxLayout()
        self.fuzzy_label = QLabel("Exact Match:")
        self.fuzzy_is_disabled = QCheckBox()

        self.fuzzy_is_disabled.stateChanged.connect(self._state_change_fuzzy_search)
        self.fuzzy_is_disabled.setToolTip(
            "Enable approximate matching (OFF) and exact matching (ON)"
        )
        self.fuzzy_label.setToolTip("Enable approximate matching (OFF) and exact matching (ON)")
        fuzzy_layout.addWidget(self.fuzzy_label)
        fuzzy_layout.addWidget(self.fuzzy_is_disabled)
        fuzzy_layout.addStretch()

        # Add both search components to the layout
        self.search_controls = QHBoxLayout()
        self.search_controls.addLayout(search_layout)
        self.search_controls.addSpacing(20)  # Add some space between the search box and toggle
        self.search_controls.addLayout(fuzzy_layout)

        # Add filter section for proposals

        filter_layout = QHBoxLayout()
        filter_layout.setContentsMargins(12, 0, 12, 0)
        filter_layout.setSpacing(12)
        self._with_proposals = QCheckBox("Show experiments with proposals", self)
        self._without_proposals = QCheckBox("Show experiments without proposals", self)
        self._with_proposals.setChecked(True)
        self._without_proposals.setChecked(True)
        self._with_proposals.toggled.connect(self._apply_table_filters)
        self._without_proposals.toggled.connect(self._apply_table_filters)
        filter_layout.addWidget(self._with_proposals)
        filter_layout.addWidget(self._without_proposals)
        filter_layout.addStretch(1)
        self.search_controls.addLayout(filter_layout)

        # Insert the search controls layout at the top of the table
        layout.addLayout(self.search_controls)

    def _build_table_tab(self):
        layout = QVBoxLayout(self._table_tab)
        layout.setContentsMargins(16, 16, 16, 16)
        layout.setSpacing(8)

        self._setup_search(layout)

        # Add table
        hor_layout = QHBoxLayout()
        self._table = QTableWidget(self._table_tab)
        self._table.setColumnCount(5)
        self._table.setHorizontalHeaderLabels(list(self._headers.values()))
        vh = self._table.verticalHeader()
        vh.setVisible(False)
        vh.setDefaultSectionSize(vh.minimumSectionSize())
        self._table.setSelectionBehavior(QTableWidget.SelectionBehavior.SelectRows)
        self._table.setSelectionMode(QTableWidget.SelectionMode.SingleSelection)
        self._table.setWordWrap(True)
        self._table.setStyleSheet("QTableWidget::item { padding: 4px; }")

        header = self._table.horizontalHeader()
        header.setSectionResizeMode(0, QHeaderView.ResizeMode.ResizeToContents)
        header.setSectionResizeMode(1, QHeaderView.ResizeMode.Stretch)
        header.setSectionResizeMode(2, QHeaderView.ResizeMode.ResizeToContents)
        header.setSectionResizeMode(3, QHeaderView.ResizeMode.ResizeToContents)
        header.setSectionResizeMode(4, QHeaderView.ResizeMode.ResizeToContents)

        self._table.itemSelectionChanged.connect(self._update_selection_state)
        hor_layout.addWidget(self._table, stretch=5)
        hor_layout.addSpacing(12)  # Add space between table and side card

        # Add side card for experiment details
        self._side_card = ExperimentMatCard(
            parent=self, show_activate_button=True, button_text="Activate"
        )
        self._side_card.experiment_selected.connect(self._emit_selected_experiment)
        hor_layout.addWidget(self._side_card, stretch=2)  # Ratio 5:2 between table and card
        layout.addLayout(hor_layout)

    @SafeSlot()
    @SafeSlot(int)
    @SafeSlot(bool)  # Overload for buttons
    def _apply_table_filters(self, *args, **kwargs):
        if self._tabs.currentWidget() is not self._table_tab:
            return

        show_with = self._with_proposals.isChecked()
        show_without = self._without_proposals.isChecked()

        self._table_infos = []
        for info in self._experiment_infos:
            has_proposal = bool(info.get("proposal"))
            if has_proposal and not show_with:
                continue
            if not has_proposal and not show_without:
                continue
            self._table_infos.append(info)

        self._populate_table()
        self._update_selection_state()

    def _populate_table(self):
        # Clear table before populating, this keeps headers intact
        self._table.setRowCount(0)
        # Refill table
        self._table.setRowCount(len(self._table_infos))
        for row, info in enumerate(self._table_infos):
            pgroup = info.get("pgroup", "")
            title = info.get("title", "")
            name = format_name(info)
            start, end = format_schedule(info.get("schedule"))

            self._table.setItem(row, 0, QTableWidgetItem(pgroup))
            self._table.setItem(row, 1, QTableWidgetItem(title))
            self._table.setItem(row, 2, QTableWidgetItem(name))
            self._table.setItem(row, 3, QTableWidgetItem(start))
            self._table.setItem(row, 4, QTableWidgetItem(end))

        width = self._table.viewport().width()
        self._table.resizeRowsToContents()
        self._table.resize(width, self._table.height())
        # self._table.resizeRowsToContents()

    @SafeSlot()
    def _update_selection_state(self):
        if self._tabs.currentWidget() is not self._table_tab:
            return
        index = self._table.selectionModel().selectedRows()
        if len(index) > 0:
            index = index[0]
            self._side_card.set_experiment_info(self._table_infos[index.row()])

    def _emit_selected_experiment(self):
        if self._tabs.currentWidget() is self._card_tab and self._next_experiment:
            self.experiment_selected.emit(self._next_experiment)
            return
        selected = self._table.selectionModel().selectedRows()
        if not selected:
            return
        row = selected[0].row()
        if 0 <= row < len(self._table_infos):
            self.experiment_selected.emit(self._table_infos[row])
            logger.info(f"Emitting next experiment signal with info: {self._table_infos[row]}")

    def _select_next_experiment(self, experiment_infos: list[dict]) -> dict | None:
        candidates = []
        for info in experiment_infos:
            start, _ = format_schedule(info.get("schedule"), as_datetime=True)
            if start is None:
                continue
            candidates.append((start, info))

        if not candidates:
            return experiment_infos[0] if experiment_infos else None

        now = datetime.now()
        future = [entry for entry in candidates if entry[0] >= now]
        pool = future or candidates
        return min(pool, key=lambda entry: abs(entry[0] - now))[1]

    def _on_tab_changed(self, index):
        if self._tabs.widget(index) is self._table_tab:
            self._table.resizeRowsToContents()
            if self._next_experiment:
                self._side_card.set_experiment_info(self._next_experiment)
        self._apply_table_filters()

    def _get_column_data(self, row) -> dict[str, str]:
        output = {}
        for ii, header in enumerate(self._headers.values()):
            item = self._table.item(row, ii)
            if item is None:
                output[header] = ""
                continue
            output[header] = item.text()
        return output

    @SafeSlot(str)
    def _apply_row_filter(self, text_input: str):
        """Apply a filter to the table rows based on the filter text."""
        if not text_input:
            for row in self._hidden_rows:
                self._table.setRowHidden(row, False)
            self._hidden_rows.clear()
            return
        for row in range(self._table.rowCount()):
            experiment_data = self._get_column_data(row)
            if is_match(
                text_input, experiment_data, list(self._headers.values()), self._enable_fuzzy_search
            ):
                self._table.setRowHidden(row, False)
                self._hidden_rows.discard(row)
            else:
                self._table.setRowHidden(row, True)
                self._hidden_rows.add(row)

    @SafeSlot(int)
    def _state_change_fuzzy_search(self, enabled: int):
        """Handle state changes for the fuzzy search toggle."""
        self._enable_fuzzy_search = not bool(enabled)
        # Re-apply filter with updated fuzzy search setting
        current_text = self.search_input.text()
        self._apply_row_filter(current_text)


if __name__ == "__main__":  # pragma: no cover
    from qtpy.QtWidgets import QApplication

    experiment_infos = [
        {
            "_id": "p22622",
            "owner_groups": ["admin"],
            "access_groups": ["unx-sls_xda_bs", "p22622"],
            "realm_id": "TestBeamline",
            "proposal": "12345967",
            "title": "Test Experiment for Mat Card Widget",
            "firstname": "John",
            "lastname": "Doe",
            "email": "john.doe@psi.ch",
            "account": "doe_j",
            "pi_firstname": "Jane",
            "pi_lastname": "Smith",
            "pi_email": "jane.smith@psi.ch",
            "pi_account": "smith_j",
            "eaccount": "e22622",
            "pgroup": "p22622",
            "abstract": "This is a test abstract for the experiment mat card widget. It should be long enough to test text wrapping and display in the card. The abstract provides a brief overview of the experiment, its goals, and its significance. This text is meant to simulate a real abstract that might be associated with an experiment in the BEC Atlas system. The card should be able to handle abstracts of varying lengths without any issues, ensuring that the user can read the full abstract even if it is quite long.",
            "schedule": [{"start": "01/01/2025 08:00:00", "end": "03/01/2025 18:00:00"}],
            "proposal_submitted": "15/12/2024",
            "proposal_expire": "31/12/2025",
            "proposal_status": "Scheduled",
            "delta_last_schedule": 30,
            "mainproposal": "",
        },
        {
            "_id": "p22623",
            "owner_groups": ["admin"],
            "access_groups": ["unx-sls_xda_bs", "p22623"],
            "realm_id": "TestBeamline",
            "proposal": "",
            "title": "Experiment without Proposal",
            "firstname": "Alice",
            "lastname": "Johnson",
            "email": "alice.johnson@psi.ch",
            "account": "johnson_a",
            "pi_firstname": "Bob",
            "pi_lastname": "Brown",
            "pi_email": "bob.brown@psi.ch",
            "pi_account": "brown_b",
            "eaccount": "e22623",
            "pgroup": "p22623",
            "abstract": "",
            "schedule": [],
            "proposal_submitted": "",
            "proposal_expire": "",
            "proposal_status": "",
            "delta_last_schedule": None,
            "mainproposal": "",
        },
    ]

    app = QApplication([])
    from bec_qthemes import apply_theme

    from bec_widgets.widgets.utility.visual.dark_mode_button.dark_mode_button import DarkModeButton

    apply_theme("light")
    w = QWidget()
    l = QVBoxLayout(w)
    dark_button = DarkModeButton()
    l.addWidget(dark_button)
    widget = ExperimentSelection(experiment_infos)
    l.addWidget(widget)
    w.resize(1280, 920)
    w.show()
    app.exec()



from __future__ import annotations

import os
import pathlib
from typing import Any, cast

from bec_lib.logger import bec_logger
from bec_lib.macro_update_handler import has_executable_code
from qtpy.QtCore import Signal
from qtpy.QtWidgets import QFileDialog, QMessageBox, QToolButton, QWidget

from bec_widgets.widgets.containers.dock_area.basic_dock_area import DockAreaWidget
from bec_widgets.widgets.containers.qt_ads import CDockAreaWidget, CDockWidget
from bec_widgets.widgets.editors.monaco.monaco_widget import MonacoWidget

logger = bec_logger.logger


class MonacoDock(DockAreaWidget):
    """
    MonacoDock is a dock widget that contains Monaco editor instances.
    It is used to manage multiple Monaco editors in a dockable interface.
    """

    focused_editor = Signal(object)  # Emitted when the focused editor changes
    save_enabled = Signal(bool)  # Emitted when the save action is enabled/disabled
    signature_help = Signal(str)  # Emitted when signature help is requested
    macro_file_updated = Signal(str)  # Emitted when a macro file is saved

    def __init__(self, parent=None, **kwargs):
        super().__init__(
            parent=parent,
            variant="compact",
            title="Monaco Editors",
            default_add_direction="top",
            **kwargs,
        )
        self.dock_manager.focusedDockWidgetChanged.connect(self._on_focus_event)
        self._last_focused_editor: CDockWidget | None = None
        self.focused_editor.connect(self._on_last_focused_editor_changed)
        initial_editor = self.add_editor()
        if isinstance(initial_editor, CDockWidget):
            self.last_focused_editor = initial_editor
        self._install_manager_scan_and_fix_guards()

    def _create_editor_widget(self) -> MonacoWidget:
        """Create a configured Monaco editor widget."""
        init_lsp = len(self.dock_manager.dockWidgets()) == 0
        widget = MonacoWidget(self, init_lsp=init_lsp)
        widget.save_enabled.connect(self.save_enabled.emit)
        widget.editor.signature_help_triggered.connect(self._on_signature_change)
        return widget

    @property
    def last_focused_editor(self) -> CDockWidget | None:
        """
        Get the last focused editor.
        """
        return self._last_focused_editor

    @last_focused_editor.setter
    def last_focused_editor(self, editor: CDockWidget | None):
        self._last_focused_editor = editor
        self.focused_editor.emit(editor)

    def _on_last_focused_editor_changed(self, editor: CDockWidget | None):
        if editor is None:
            self.save_enabled.emit(False)
            return

        widget = cast(MonacoWidget, editor.widget())
        if widget.modified:
            logger.info(f"Editor '{widget.current_file}' has unsaved changes: {widget.get_text()}")
        self.save_enabled.emit(widget.modified)

    @staticmethod
    def _update_tab_title_for_modification(dock: CDockWidget, modified: bool):
        """Update the tab title to show modification status with a dot indicator."""
        current_title = dock.windowTitle()

        # Remove existing modification indicator (dot and space)
        if current_title.startswith("• "):
            base_title = current_title[2:]  # Remove "• "
        else:
            base_title = current_title

        # Add or remove the modification indicator
        if modified:
            new_title = f"• {base_title}"
        else:
            new_title = base_title

        dock.setWindowTitle(new_title)

    def _on_signature_change(self, signature: dict):
        signatures = signature.get("signatures", [])
        if not signatures:
            self.signature_help.emit("")
            return

        active_sig = signatures[signature.get("activeSignature", 0)]
        # Get signature label and documentation
        label = active_sig.get("label", "")
        doc_obj = active_sig.get("documentation", {})
        documentation = doc_obj.get("value", "") if isinstance(doc_obj, dict) else str(doc_obj)

        # Format the Markdown output
        markdown = f"```python\n{label}\n```\n\n{documentation}"
        self.signature_help.emit(markdown)

    def _on_focus_event(self, old_widget, new_widget) -> None:
        # Track focus events for the dock widget
        widget = new_widget.widget()
        if isinstance(widget, MonacoWidget):
            self.last_focused_editor = new_widget

    def _on_editor_close_requested(self, dock: CDockWidget, widget: QWidget):
        # Cast widget to MonacoWidget since we know that's what it is
        monaco_widget = cast(MonacoWidget, widget)

        # Check if we have unsaved changes
        if monaco_widget.modified:
            # Prompt the user to save changes
            response = QMessageBox.question(
                self,
                "Unsaved Changes",
                "You have unsaved changes. Do you want to save them?",
                QMessageBox.StandardButton.Yes
                | QMessageBox.StandardButton.No
                | QMessageBox.StandardButton.Cancel,
            )
            if response == QMessageBox.StandardButton.Yes:
                self.save_file(monaco_widget)
            elif response == QMessageBox.StandardButton.Cancel:
                return

        # Count all editor docks managed by this dock manager
        total = len(self.dock_manager.dockWidgets())
        if total <= 1:
            # Do not remove the last dock; just wipe its editor content
            # Temporarily disable read-only mode if the editor is read-only
            # so we can clear the content for reuse
            self.reset_widget(monaco_widget)
            dock.setWindowTitle("Untitled")
            dock.setTabToolTip("Untitled")
            icon = self._resolve_dock_icon(monaco_widget, dock_icon=None, apply_widget_icon=True)
            dock.setIcon(icon)
            self.last_focused_editor = dock
            return

        # Otherwise, proceed to close and delete the dock
        monaco_widget.close()
        dock.closeDockWidget()
        dock.deleteDockWidget()
        if self.last_focused_editor is dock:
            self.last_focused_editor = None
        # After topology changes, make sure single-tab areas get a plus button
        self._scan_and_fix_areas()

    @staticmethod
    def reset_widget(widget: MonacoWidget):
        """
        Reset the given Monaco editor widget to its initial state.

        Args:
            widget (MonacoWidget): The Monaco editor widget to reset.
        """
        widget.set_readonly(False)
        widget.set_text("", reset=True)
        widget.metadata["scope"] = ""

    def _ensure_area_plus(self, area):
        if area is None:
            return
        # Only add once per area
        if getattr(area, "_monaco_plus_btn", None) is not None:
            return
        # If the area has exactly one tab, inject a + button next to the tab bar
        try:
            tabbar = area.titleBar().tabBar()
            count = tabbar.count() if hasattr(tabbar, "count") else 1
        except Exception:
            count = 1
        if count >= 1:
            plus_btn = QToolButton(area)
            plus_btn.setText("+")
            plus_btn.setToolTip("New Monaco Editor")
            plus_btn.setAutoRaise(True)
            tb = area.titleBar()
            idx = tb.indexOf(tb.tabBar())
            tb.insertWidget(idx + 1, plus_btn)
            plus_btn.clicked.connect(lambda: self.add_editor(area))
            # pylint: disable=protected-access
            area._monaco_plus_btn = plus_btn

    def _install_manager_scan_and_fix_guards(self) -> None:
        """
        Track ADS structural changes to trigger scan and fix of dock areas for plus button injection.
        """
        self.dock_manager.dockAreaCreated.connect(self._scan_and_fix_areas)
        self.dock_manager.dockWidgetAdded.connect(self._scan_and_fix_areas)
        self.dock_manager.stateRestored.connect(self._scan_and_fix_areas)
        self.dock_manager.restoringState.connect(self._scan_and_fix_areas)
        self.dock_manager.focusedDockWidgetChanged.connect(self._scan_and_fix_areas)
        self._scan_and_fix_areas()

    def _scan_and_fix_areas(self, *_arg):
        # Find all dock areas under this manager and ensure each single-tab area has a plus button
        areas = self.dock_manager.findChildren(CDockAreaWidget)
        for a in areas:
            self._ensure_area_plus(a)

    def add_editor(
        self, area: Any | None = None, title: str | None = None, tooltip: str | None = None
    ) -> CDockWidget:
        """
        Add a new Monaco editor dock to the specified area.

        Args:
            area(Any | None): The area to add the editor to. If None, adds to the main area.
            title(str | None): The title of the editor tab. If None, a default title is used.
            tooltip(str | None): The tooltip for the editor tab. If None, no tooltip is set.

        Returns:
            CDockWidget: The created dock widget containing the Monaco editor.
        """
        widget = self._create_editor_widget()
        existing_count = len(self.dock_manager.dockWidgets())
        default_title = title or f"Untitled_{existing_count + 1}"

        tab_target: CDockWidget | None = None
        if isinstance(area, CDockAreaWidget):
            tab_target = area.currentDockWidget()
            if tab_target is None:
                docks = area.dockWidgets()
                tab_target = docks[0] if docks else None

        dock = self.new(
            widget,
            closable=True,
            floatable=False,
            movable=True,
            tab_with=tab_target,
            return_dock=True,
            on_close=self._on_editor_close_requested,
            title_buttons={"float": False},
            where="right",
        )
        dock.setWindowTitle(default_title)
        if tooltip is not None:
            dock.setTabToolTip(tooltip)

        widget.save_enabled.connect(
            lambda modified, target=dock: self._update_tab_title_for_modification(target, modified)
        )

        area_widget = dock.dockAreaWidget()
        if area_widget is not None:
            self._ensure_area_plus(area_widget)

        self._scan_and_fix_areas()
        self.last_focused_editor = dock
        return dock

    def open_file(self, file_name: str, scope: str = "") -> None:
        """
        Open a file in the specified area. If the file is already open, activate it.

        Args:
            file_name (str): The path to the file to open.
            scope (str): The scope to set for the editor metadata.
        """

        open_files = self._get_open_files()
        if file_name in open_files:
            dock = self._get_editor_dock(file_name)
            if dock is not None:
                dock.setAsCurrentTab()
                self.last_focused_editor = dock
            return

        file = os.path.basename(file_name)
        # If the current editor is empty, we reuse it

        # For now, the dock manager is only for the editor docks. We can therefore safely assume
        # that all docks are editor docks.
        dock_area = self.dock_manager.dockArea(0)
        if not dock_area:
            return

        editor_dock = dock_area.currentDockWidget()
        if not editor_dock:
            return

        editor_widget = editor_dock.widget() if editor_dock else None
        if editor_widget:
            editor_widget = cast(MonacoWidget, editor_dock.widget())
            if editor_widget.current_file is None and editor_widget.get_text() == "":
                editor_dock.setWindowTitle(file)
                editor_dock.setTabToolTip(file_name)
                editor_widget.open_file(file_name)
                editor_widget.metadata["scope"] = scope
                self.last_focused_editor = editor_dock
                return

        # File is not open, create a new editor
        editor_dock = self.add_editor(title=file, tooltip=file_name)
        widget = cast(MonacoWidget, editor_dock.widget())
        widget.open_file(file_name)
        widget.metadata["scope"] = scope
        editor_dock.setAsCurrentTab()
        self.last_focused_editor = editor_dock

    def save_file(
        self, widget: MonacoWidget | None = None, force_save_as: bool = False, format_on_save=True
    ) -> None:
        """
        Save the currently focused file.

        Args:
            widget (MonacoWidget | None): The widget to save. If None, the last focused editor will be used.
            force_save_as (bool): If True, the "Save As" dialog will be shown even if the file is already saved.
            format_on_save (bool): If True, format the code before saving if it's a Python file.
        """
        if widget is None:
            widget = self.last_focused_editor.widget() if self.last_focused_editor else None
        if not widget:
            return
        if "macros" in widget.metadata.get("scope", ""):
            if not self._validate_macros(widget.get_text()):
                return

        if widget.current_file and not force_save_as:
            if format_on_save and pathlib.Path(widget.current_file).suffix == ".py":
                widget.format()

            with open(widget.current_file, "w", encoding="utf-8") as f:
                f.write(widget.get_text())

            if "macros" in widget.metadata.get("scope", ""):
                self._update_macros(widget)
                # Emit signal to refresh macro tree widget
                self.macro_file_updated.emit(widget.current_file)

            # pylint: disable=protected-access
            widget._original_content = widget.get_text()
            widget.save_enabled.emit(False)
            return

        # Save as option
        save_file = QFileDialog.getSaveFileName(self, "Save File As", "", "All files (*)")

        if not save_file or not save_file[0]:
            return
        # check if we have suffix specified
        file = pathlib.Path(save_file[0])
        if file.suffix == "":
            file = file.with_suffix(".py")
        if format_on_save and file.suffix == ".py":
            widget.format()

        text = widget.get_text()
        with open(file, "w", encoding="utf-8") as f:
            f.write(text)
        widget._original_content = text

        # Update the current_file before emitting save_enabled to ensure proper tracking
        widget._current_file = str(file)
        widget.save_enabled.emit(False)

        # Find the dock widget containing this monaco widget and update title
        for dock in self.dock_manager.dockWidgets():
            if dock.widget() == widget:
                dock.setWindowTitle(file.name)
                dock.setTabToolTip(str(file))
                break
        if "macros" in widget.metadata.get("scope", ""):
            self._update_macros(widget)
            # Emit signal to refresh macro tree widget
            self.macro_file_updated.emit(str(file))

        logger.debug(f"Save file called, last focused editor: {self.last_focused_editor}")

    def _validate_macros(self, source: str) -> bool:
        # pylint: disable=protected-access
        # Ensure the macro does not contain executable code before saving
        exec_code, line_number = has_executable_code(source)
        if exec_code:
            if line_number is None:
                msg = "The macro contains executable code. Please remove it before saving."
            else:
                msg = f"The macro contains executable code on line {line_number}. Please remove it before saving."
            QMessageBox.warning(self, "Save Error", msg)
            return False
        return True

    def _update_macros(self, widget: MonacoWidget):
        # pylint: disable=protected-access
        if not widget.current_file:
            return
        # Check which macros have changed and broadcast the change
        macros = self.client.macros._update_handler.get_macros_from_file(widget.current_file)
        existing_macros = self.client.macros._update_handler.get_existing_macros(
            widget.current_file
        )

        removed_macros = set(existing_macros.keys()) - set(macros.keys())
        added_macros = set(macros.keys()) - set(existing_macros.keys())
        for name, info in macros.items():
            if name in added_macros:
                self.client.macros._update_handler.broadcast(
                    action="add", name=name, file_path=widget.current_file
                )
            if (
                name in existing_macros
                and info.get("source", "") != existing_macros[name]["source"]
            ):
                self.client.macros._update_handler.broadcast(
                    action="reload", name=name, file_path=widget.current_file
                )
        for name in removed_macros:
            self.client.macros._update_handler.broadcast(action="remove", name=name)

    def set_vim_mode(self, enabled: bool):
        """
        Set Vim mode for all editor widgets.

        Args:
            enabled (bool): Whether to enable or disable Vim mode.
        """
        for widget in self.dock_manager.dockWidgets():
            editor_widget = cast(MonacoWidget, widget.widget())
            editor_widget.set_vim_mode_enabled(enabled)

    def _get_open_files(self) -> list[str]:
        open_files = []
        for widget in self.dock_manager.dockWidgets():
            editor_widget = cast(MonacoWidget, widget.widget())
            if editor_widget.current_file is not None:
                open_files.append(editor_widget.current_file)
        return open_files

    def _get_editor_dock(self, file_name: str) -> CDockWidget | None:
        for widget in self.dock_manager.dockWidgets():
            editor_widget = cast(MonacoWidget, widget.widget())
            if editor_widget.current_file == file_name:
                return widget
        return None

    def set_file_readonly(self, file_name: str, read_only: bool = True) -> bool:
        """
        Set a specific file's editor to read-only mode.

        Args:
            file_name (str): The file path to set read-only
            read_only (bool): Whether to set read-only mode (default: True)

        Returns:
            bool: True if the file was found and read-only was set, False otherwise
        """
        editor_dock = self._get_editor_dock(file_name)
        if editor_dock:
            editor_widget = cast(MonacoWidget, editor_dock.widget())
            editor_widget.set_readonly(read_only)
            return True
        return False

    def set_file_icon(self, file_name: str, icon) -> bool:
        """
        Set an icon for a specific file's tab.

        Args:
            file_name (str): The file path to set icon for
            icon: The QIcon to set on the tab

        Returns:
            bool: True if the file was found and icon was set, False otherwise
        """
        editor_dock = self._get_editor_dock(file_name)
        if editor_dock:
            editor_dock.setIcon(icon)
            return True
        return False


if __name__ == "__main__":
    import sys

    from qtpy.QtWidgets import QApplication

    app = QApplication(sys.argv)
    _dock = MonacoDock()
    _dock.show()
    sys.exit(app.exec())



"""
Module for a TableWidget for the device manager view. Row data is encapsulated
in DeviceTableRow entries.
"""

from __future__ import annotations

from copy import deepcopy
from typing import TYPE_CHECKING, Any, Callable, Iterable, Tuple

from bec_lib.atlas_models import Device as DeviceModel
from bec_lib.callback_handler import EventType
from bec_lib.logger import bec_logger
from bec_qthemes import material_icon
from qtpy import QtCore, QtGui, QtWidgets
from thefuzz import fuzz

from bec_widgets.utils.bec_widget import BECWidget
from bec_widgets.utils.colors import get_accent_colors
from bec_widgets.utils.error_popups import SafeSlot
from bec_widgets.utils.fuzzy_search import is_match
from bec_widgets.widgets.control.device_manager.components.device_table.device_table_row import (
    DeviceTableRow,
)
from bec_widgets.widgets.control.device_manager.components.ophyd_validation import (
    ConfigStatus,
    ConnectionStatus,
    get_validation_icons,
)

if TYPE_CHECKING:  # pragma: no cover
    from bec_lib.messages import ConfigAction

logger = bec_logger.logger

_DeviceCfgIter = Iterable[dict[str, Any]]
# DeviceValidationResult: device_config, config_status, connection_status, error_message
_ValidationResultIter = Iterable[Tuple[dict[str, Any], ConfigStatus, ConnectionStatus, str]]


class TableSortOnHold:
    """Context manager for putting table sorting on hold. Works with nested calls."""

    def __init__(self, table: QtWidgets.QTableWidget) -> None:
        self.table = table
        self._call_depth = 0
        self._registered_methods = []

    def register_on_hold_method(
        self, method: Callable[[QtWidgets.QTableWidget, bool], None]
    ) -> None:
        """
        Register a method to be called when sorting is put on hold.

        Args:
            method (Callable[[QtWidgets.QTableWidget, bool], None]): The method to register.
                            The method should accept the QTableWidget and a bool indicating
                            whether sorting is being enabled (True) or disabled (False).
        """
        self._registered_methods.append(method)

    def __enter__(self):
        """Enter the context manager"""
        self._call_depth += 1  # Needed for nested calls
        self.table.setSortingEnabled(False)
        for method in self._registered_methods:
            method(self.table, False)

    def __exit__(self, *exc):
        """Exit the context manager"""
        self._call_depth -= 1  # Remove nested calls
        if self._call_depth == 0:  # Only re-enable sorting on outermost exit
            self.table.setSortingEnabled(True)
            for method in self._registered_methods:
                method(self.table, True)


class CenterIconDelegate(QtWidgets.QStyledItemDelegate):
    """Custom delegate to center icons in table cells."""

    def paint(
        self,
        painter: QtGui.QPainter,
        option: QtWidgets.QStyleOptionViewItem,
        index: QtCore.QModelIndex,
    ):
        # First draw the default cell (without icon)
        opt = QtWidgets.QStyleOptionViewItem(option)
        self.initStyleOption(opt, index)
        opt.icon = QtGui.QIcon()  # Create empty icon to avoid default to be drawn at given position
        option.widget.style().drawControl(
            QtWidgets.QStyle.ControlElement.CE_ItemViewItem, opt, painter, option.widget
        )
        # Check if there is an icon to draw
        icon = index.data(QtCore.Qt.ItemDataRole.DecorationRole)
        if not icon:
            return
        # Draw the icon centered in the cell
        icon_size = option.decorationSize
        if icon_size.isValid():
            size = icon_size
        else:
            size = icon.actualSize(option.rect.size())

        x = option.rect.x() + (option.rect.width() - size.width()) // 2
        y = option.rect.y() + (option.rect.height() - size.height()) // 2

        icon.paint(painter, QtCore.QRect(QtCore.QPoint(x, y), size))


class CheckBoxDelegate(QtWidgets.QStyledItemDelegate):
    """Custom delegate to handle checkbox interactions in the table."""

    # Signal to indicate a checkbox was clicked
    checkbox_clicked = QtCore.Signal(int, int, bool)  # row, column, checked

    def editorEvent(
        self,
        event: QtCore.QEvent,
        model: QtCore.QAbstractItemModel,
        option: QtWidgets.QStyleOptionViewItem,
        index: QtCore.QModelIndex,
    ):
        if event.type() == QtCore.QEvent.Type.MouseButtonRelease:
            if model and (model.flags(index) & QtCore.Qt.ItemFlag.ItemIsUserCheckable):
                old_state = QtCore.Qt.CheckState(
                    model.data(index, QtCore.Qt.ItemDataRole.CheckStateRole)
                )
                new_state = (
                    QtCore.Qt.CheckState.Unchecked
                    if old_state == QtCore.Qt.CheckState.Checked
                    else QtCore.Qt.CheckState.Checked
                )
                model.setData(index, new_state, QtCore.Qt.ItemDataRole.CheckStateRole)
                model.setData(
                    index,
                    new_state == QtCore.Qt.CheckState.Checked,
                    QtCore.Qt.ItemDataRole.UserRole,
                )
                self.checkbox_clicked.emit(
                    index.row(), index.column(), new_state == QtCore.Qt.CheckState.Checked
                )
            return True
        return super().editorEvent(event, model, option, index)


class SortTableItem(QtWidgets.QTableWidgetItem):
    """Custom TableWidgetItem with hidden __column_data attribute for sorting."""

    def __lt__(self, other: QtWidgets.QTableWidgetItem) -> bool:
        """Override less-than operator for sorting."""
        if not isinstance(other, QtWidgets.QTableWidgetItem):
            return NotImplemented
        self_data = self.data(QtCore.Qt.ItemDataRole.UserRole)
        other_data = other.data(QtCore.Qt.ItemDataRole.UserRole)
        if self_data is not None and other_data is not None:
            return self_data < other_data
        return super().__lt__(other)

    def __gt__(self, other: QtWidgets.QTableWidgetItem) -> bool:
        """Override less-than operator for sorting."""
        if not isinstance(other, QtWidgets.QTableWidgetItem):
            return NotImplemented
        self_data = self.data(QtCore.Qt.ItemDataRole.UserRole)
        other_data = other.data(QtCore.Qt.ItemDataRole.UserRole)
        if self_data is not None and other_data is not None:
            return self_data > other_data
        return super().__gt__(other)


class DeviceTable(BECWidget, QtWidgets.QWidget):
    """Custom table to display device configurations."""

    RPC = False  # TODO discuss if this should be available for RPC

    # Signal emitted if devices are added (updated) or removed
    #   - device_configs: List of device configurations.
    #   - added: True if devices were added/updated, False if removed.
    #   - skip validation: True if validation should be skipped for added/updated devices.
    device_configs_changed = QtCore.Signal(list, bool, bool)
    # Signal emitted when device selection changes, emits list of selected device configs
    selected_devices = QtCore.Signal(list)
    # Signal emitted when a device row is double-clicked, emits the device config
    device_row_dbl_clicked = QtCore.Signal(dict)
    # Signal emitted when the device config is in sync with Redis
    device_config_in_sync_with_redis = QtCore.Signal(bool)

    # Request multiple validation updates for devices
    request_update_multiple_device_validations = QtCore.Signal(list)
    # Request update after client DEVICE_UPDATE event
    request_update_after_client_device_update = QtCore.Signal()

    _auto_size_request = QtCore.Signal()

    def __init__(self, parent: QtWidgets.QWidget | None = None, client=None):
        super().__init__(parent=parent, client=client)
        self.headers_key_map: dict[str, str] = {
            "Valid": "valid",
            "Connect": "connect",
            "Name": "name",
            "Device Class": "deviceClass",
            "Readout Priority": "readoutPriority",
            "On Failure": "onFailure",
            "Device Tags": "deviceTags",
            "Description": "description",
            "Enabled": "enabled",
            "Read Only": "readOnly",
            "Software Trigger": "softwareTrigger",
        }

        # General attributes
        self._icon_size = (18, 18)
        self._colors = get_accent_colors()
        self._icons = get_validation_icons(self._colors, self._icon_size)
        self._check_box_icons = {
            "checked": material_icon(
                "check_box", size=(24, 24), color=self._colors.default, convert_to_pixmap=False
            ),
            "unchecked": material_icon(
                "check_box_outline_blank",
                size=(24, 24),
                color=self._colors.default,
                convert_to_pixmap=False,
            ),
        }
        self._layout = QtWidgets.QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(4)
        self.setLayout(self._layout)

        # Table related attributes
        self.row_data: dict[str, DeviceTableRow] = {}
        self.table = QtWidgets.QTableWidget(self)
        self.table_sort_on_hold = TableSortOnHold(self.table)
        self._setup_table()
        self.table_sort_on_hold.register_on_hold_method(self._resize_table_policy)
        self.table_sort_on_hold.register_on_hold_method(self._set_table_signals_on_hold)

        # Search related attributes
        self._searchable_keys: list[str] = ["name", "deviceClass", "deviceTags", "description"]
        self._hidden_rows: set[int] = set()
        self._enable_fuzzy_search: bool = True
        self._setup_search()

        # Add components to layout
        self._layout.addLayout(self.search_controls)
        self._layout.addWidget(self.table)

        # Connect slots
        self.table.selectionModel().selectionChanged.connect(self._on_selection_changed)
        self.table.cellDoubleClicked.connect(self._on_cell_double_clicked)
        self.request_update_multiple_device_validations.connect(
            self.update_multiple_device_validations
        )
        self.request_update_after_client_device_update.connect(self._on_device_config_update)
        # Install event filter
        self.table.installEventFilter(self)

        # Add hook to BECClient for DeviceUpdates
        self.client_callback_id = self.client.callbacks.register(
            event_type=EventType.DEVICE_UPDATE, callback=self.__on_client_device_update_event
        )

    def cleanup(self):
        """Cleanup resources."""
        self.row_data.clear()  # Drop references to row data..
        self.client.callbacks.remove(self.client_callback_id)  # Unregister callback
        super().cleanup()

    def __on_client_device_update_event(
        self, action: "ConfigAction", config: dict[str, dict[str, Any]]
    ) -> None:
        """Handle DEVICE_UPDATE events from the BECClient."""
        self.request_update_after_client_device_update.emit()

    @SafeSlot()
    def _on_device_config_update(self) -> None:
        """Handle device configuration updates from the BECClient."""
        # Determine the overlapping device configs between Redis and the table
        device_config_overlap_with_bec = self._get_overlapping_configs()
        if len(device_config_overlap_with_bec) > 0:
            # Notify any listeners about the update, the device manager devices will now be up to date
            self.device_configs_changed.emit(device_config_overlap_with_bec, True, True)

        # Correct all connection statuses in the table which are ConnectionStatus.CONNECTED
        # to ConnectionStatus.CAN_CONNECT
        device_status_updates = []
        validation_results = self.get_validation_results()
        for device_name, (cfg, config_status, connection_status) in validation_results.items():
            if device_name is None:
                continue
            # Check if config is not in the overlap, but connection status is CONNECTED
            # Update to CAN_CONNECT
            if cfg not in device_config_overlap_with_bec:
                if connection_status == ConnectionStatus.CONNECTED.value:
                    device_status_updates.append(
                        (cfg, config_status, ConnectionStatus.CAN_CONNECT.value, "")
                    )
        # Update only if there are any updates
        if len(device_status_updates) > 0:
            # NOTE We need to emit here a signal to call update_multiple_device_validations
            # as this otherwise can cause problems with being executed from a python callback
            # thread which are not properly scheduled in the Qt event loop. We see that this
            # has caused issues in form of segfaults under certain usage of the UI. Please
            # do not remove this signal & slot mechanism!
            self.request_update_multiple_device_validations.emit(device_status_updates)

        # Check if in sync with BEC server session
        in_sync_with_redis = self._is_config_in_sync_with_redis()
        self.device_config_in_sync_with_redis.emit(in_sync_with_redis)

    # -------------------------------------------------------------------------
    # Custom hooks for table events
    # -------------------------------------------------------------------------

    def _on_selection_changed(
        self, selected: QtCore.QItemSelection, deselected: QtCore.QItemSelection
    ):
        """Handle selection changes in the table."""
        rows = set()
        for index in selected.indexes():
            row = index.row()
            rows.add(row)
        selected_configs = []
        for row in rows:
            device_name = self._get_cell_data(row, 2)  # Name column
            if device_name:
                row_data = self.row_data.get(device_name)
                if row_data:
                    cfg = deepcopy(row_data.data)
                    cfg.pop("name")
                    selected_configs.append({device_name: cfg})
        self.selected_devices.emit(selected_configs)

    def _on_cell_double_clicked(self, row: int, column: int):
        """Handle double-click events on table cells."""
        device_name = self._get_cell_data(row, 2)  # Name column
        if device_name:
            row_data = self.row_data.get(device_name)
            if row_data:
                self.device_row_dbl_clicked.emit(row_data.data)

    def eventFilter(self, source: QtCore.QObject, event: QtCore.QEvent) -> bool:
        """Customize event filtering for table interactions."""
        if source is self.table:
            if event.type() == QtCore.QEvent.Type.KeyPress:
                if event.key() in (QtCore.Qt.Key.Key_Backspace, QtCore.Qt.Key.Key_Delete):
                    configs = self.get_selected_device_configs()
                    if configs:
                        if self._remove_configs_dialog([cfg["name"] for cfg in configs]):
                            self.remove_device_configs(configs)
                        return True  # Event handled
                if event.key() == QtCore.Qt.Key.Key_Escape:
                    self.table.clearSelection()
                    return True  # handled
        return super().eventFilter(source, event)

    def _on_table_checkbox_clicked(self, row: int, column: int, checked: bool):
        """Handle checkbox clicks in the table."""
        name_index = list(self.headers_key_map.values()).index("name")
        device_name = self._get_cell_data(row, name_index)
        row_data = self.row_data.get(device_name)
        if not row_data:
            return
        row_data.data[self.headers_key_map[list(self.headers_key_map.keys())[column]]] = checked
        self._on_device_row_data_changed(row_data.data)

    def _on_device_row_data_changed(self, data: dict):
        """Handle data change events from device rows."""
        device_name = data.get("name", None)
        cfg = deepcopy(data)
        cfg.pop("name")
        self.selected_devices.emit([{device_name: cfg}])
        self.device_config_in_sync_with_redis.emit(self._is_config_in_sync_with_redis())

    def _apply_row_filter(self, text_input: str):
        """Apply a filter to the table rows based on the filter text."""
        for row in range(self.table.rowCount()):
            device_name = self._get_cell_data(row, 2)  # Name column
            if not device_name:
                continue
            row_data = self.row_data.get(device_name)
            if not row_data:
                continue
            if is_match(
                text_input, row_data.data, self._searchable_keys, self._enable_fuzzy_search
            ):
                self.table.setRowHidden(row, False)
                self._hidden_rows.discard(row)
            else:
                self.table.setRowHidden(row, True)
                self._hidden_rows.add(row)

    def _state_change_fuzzy_search(self, enabled: int):
        """Handle state changes for the fuzzy search toggle."""
        self._enable_fuzzy_search = not bool(enabled)
        # Re-apply filter with updated fuzzy search setting
        current_text = self.search_input.text()
        self._apply_row_filter(current_text)

    # -------------------------------------------------------------------------
    # Custom Dialog
    # -------------------------------------------------------------------------

    def _remove_configs_dialog(self, device_names: list[str]) -> bool:
        """
        Prompt the user to confirm removal of rows and remove them from the model if accepted.

        Args:
            device_names (list[str]): List of device names to be removed.

        Returns:
            bool: True if the user confirmed removal, False otherwise.
        """
        msg = QtWidgets.QMessageBox(self)
        msg.setIcon(QtWidgets.QMessageBox.Icon.Warning)
        msg.setWindowTitle("Confirm device removal")
        msg.setText(
            f"Remove device '{device_names[0]}'?"
            if len(device_names) == 1
            else f"Remove {len(device_names)} devices?"
        )
        separator = "\n" if len(device_names) < 12 else ", "
        msg.setInformativeText("Selected devices: \n" + separator.join(device_names))
        msg.setStandardButtons(
            QtWidgets.QMessageBox.StandardButton.Ok | QtWidgets.QMessageBox.StandardButton.Cancel
        )
        msg.setDefaultButton(QtWidgets.QMessageBox.StandardButton.Cancel)

        res = msg.exec_()
        if res == QtWidgets.QMessageBox.StandardButton.Ok:
            return True
        return False

    # -------------------------------------------------------------------------
    # Setup table
    # -------------------------------------------------------------------------
    def _setup_table(self):
        """Initializes the table configuration and headers."""
        # Temporary instance to get headers dynamically
        headers = list(self.headers_key_map.keys())
        self.table.setColumnCount(len(headers))
        self.table.setHorizontalHeaderLabels(headers)
        # Smooth scrolling
        self.table.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.table.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
        self.table.setEditTriggers(QtWidgets.QAbstractItemView.EditTrigger.NoEditTriggers)

        # Hide vertical header
        self.table.verticalHeader().setVisible(False)

        # Column resize policies
        header = self.table.horizontalHeader()
        header.setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeMode.Fixed)
        header.setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeMode.Fixed)
        header.setSectionResizeMode(2, QtWidgets.QHeaderView.ResizeMode.Interactive)
        header.setSectionResizeMode(3, QtWidgets.QHeaderView.ResizeMode.Interactive)
        header.setSectionResizeMode(4, QtWidgets.QHeaderView.ResizeMode.Fixed)
        header.setSectionResizeMode(5, QtWidgets.QHeaderView.ResizeMode.Fixed)
        header.setSectionResizeMode(6, QtWidgets.QHeaderView.ResizeMode.Interactive)
        header.setSectionResizeMode(7, QtWidgets.QHeaderView.ResizeMode.Stretch)
        header.setSectionResizeMode(8, QtWidgets.QHeaderView.ResizeMode.Fixed)
        header.setSectionResizeMode(9, QtWidgets.QHeaderView.ResizeMode.Fixed)
        header.setSectionResizeMode(10, QtWidgets.QHeaderView.ResizeMode.Fixed)

        for sizes, col in [
            (0, 85),
            (1, 85),
            (2, 200),
            (3, 200),
            (6, 200),
            (7, 200),
            (8, 90),
            (9, 90),
            (10, 120),
        ]:
            self.table.setColumnWidth(sizes, col)

        # Ensure column widths stay fixed
        header.setStretchLastSection(False)

        # Sorting
        self.table.setSortingEnabled(True)
        header.setSortIndicatorShown(True)
        header.setSortIndicator(2, QtCore.Qt.SortOrder.AscendingOrder)  # Default sort by name

        # Selection behavior
        self.table.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectionBehavior.SelectRows)
        self.table.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.ExtendedSelection)
        # Connect to selection model to get selection changes
        self.table.selectionModel().selectionChanged.connect(self._on_selection_changed)
        self.table.horizontalHeader().setHighlightSections(False)

        # Set delegate for checkboxes
        checkbox_delegate = CheckBoxDelegate(self.table)
        icon_delegate = CenterIconDelegate(self.table)
        self.table.setItemDelegateForColumn(0, icon_delegate)  # Config status
        self.table.setItemDelegateForColumn(1, icon_delegate)  # Connection status
        self.table.setWordWrap(True)
        for col in (8, 9, 10):  # enabled, readOnly, softwareTrigger
            self.table.setItemDelegateForColumn(col, checkbox_delegate)
        checkbox_delegate.checkbox_clicked.connect(self._on_table_checkbox_clicked)

    def _set_table_signals_on_hold(self, table: QtWidgets.QTableWidget, enable: bool):
        """Enable or disable table signals."""
        if enable:
            table.blockSignals(False)
        else:
            table.blockSignals(True)

    def _resize_table_policy(self, table: QtWidgets.QTableWidget, enable: bool):
        """Enable or disable column resizing."""
        if enable:
            table.resizeColumnToContents(2)  # Name
            table.resizeColumnToContents(3)  # Device Class
            # table.resizeRowsToContents()

    def _setup_search(self):
        """Create components related to the search functionality"""

        # Create search bar
        self.search_layout = QtWidgets.QHBoxLayout()
        self.search_label = QtWidgets.QLabel("Search:")
        self.search_input = QtWidgets.QLineEdit()
        self.search_input.setPlaceholderText("Filter devices (approximate matching)...")
        self.search_input.setClearButtonEnabled(True)
        self.search_input.textChanged.connect(self._apply_row_filter)
        self.search_layout.addWidget(self.search_label)
        self.search_layout.addWidget(self.search_input)

        # Add exact match toggle
        self.fuzzy_layout = QtWidgets.QHBoxLayout()
        self.fuzzy_label = QtWidgets.QLabel("Exact Match:")
        self.fuzzy_is_disabled = QtWidgets.QCheckBox()

        self.fuzzy_is_disabled.stateChanged.connect(self._state_change_fuzzy_search)
        self.fuzzy_is_disabled.setToolTip(
            "Enable approximate matching (OFF) and exact matching (ON)"
        )
        self.fuzzy_label.setToolTip("Enable approximate matching (OFF) and exact matching (ON)")
        self.fuzzy_layout.addWidget(self.fuzzy_label)
        self.fuzzy_layout.addWidget(self.fuzzy_is_disabled)
        self.fuzzy_layout.addStretch()

        # Add both search components to the layout
        self.search_controls = QtWidgets.QHBoxLayout()
        self.search_controls.addLayout(self.search_layout)
        self.search_controls.addSpacing(20)  # Add some space between the search box and toggle
        self.search_controls.addLayout(self.fuzzy_layout)
        QtCore.QTimer.singleShot(0, lambda: self.fuzzy_is_disabled.stateChanged.emit(0))

    # -------------------------------------------------------------------------
    # Row Management, internal methods.
    # -------------------------------------------------------------------------

    def _add_row(
        self,
        data: dict,
        config_status: ConfigStatus | int,
        connection_status: ConnectionStatus | int,
    ):
        """
        Adds a new row at the bottom and populates it with data. The row widgets
        are stored in self.row_widgets for easy access. Consider to disable sorting
        when adding rows as this method is not responsible for maintaining sort order.

        Args:
            data (dict): The device data to populate the row.
            config_status (ConfigStatus | int): The configuration validation status.
            connection_status (ConnectionStatus | int): The connection status.
        """
        with self.table_sort_on_hold:
            if data["name"] in self.row_data:
                logger.warning(f"Overwriting existing device row for {data['name']}")
                self._remove_rows_by_name([data["name"]])
            row_index = self.table.rowCount()
            self.table.insertRow(row_index)

            # Create row for the table
            device_row = DeviceTableRow(data=data)
            device_row.set_validation_status(config_status, connection_status)

            # Populate cells
            self._populate_device_row_cells(row_index, device_row)

    def _populate_device_row_cells(self, row: int, device_row: DeviceTableRow):
        """Populate the cells of a given row with the widgets from the DeviceTableRow."""
        with self.table_sort_on_hold:
            config_status, connect_status = device_row.validation_status
            column_keys = list(self.headers_key_map.values())
            for ii, key in enumerate(column_keys):
                if key in ("enabled", "readOnly", "softwareTrigger"):  # flags for checkboxes
                    item = SortTableItem()
                    item.setFlags(
                        item.flags()
                        | QtCore.Qt.ItemFlag.ItemIsUserCheckable
                        | QtCore.Qt.ItemFlag.ItemIsEnabled
                    )
                    item.setTextAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)
                elif key in ("valid", "connect"):  # status columns
                    item = SortTableItem()
                    item.setTextAlignment(
                        QtCore.Qt.AlignmentFlag.AlignCenter | QtCore.Qt.AlignmentFlag.AlignVCenter
                    )
                    item.setIcon(
                        self._icons["connection_status"][connect_status]
                        if key == "connect"
                        else self._icons["config_status"][config_status]
                    )
                else:
                    item = QtWidgets.QTableWidgetItem()
                    item.setTextAlignment(
                        QtCore.Qt.AlignmentFlag.AlignLeft | QtCore.Qt.AlignmentFlag.AlignVCenter
                    )
                self.table.setItem(row, ii, item)  # +2 offset for status columns
            self.__update_device_row_data(row, device_row.data)

    def __update_device_row_data(self, row: int, data: dict):
        """
        Update an existing device row with new data.

        Args:
            row (int): The row index to update.
            data (dict): The device data to populate the row.
        """
        # Update stored row data
        if data["name"] in self.row_data:
            self.row_data[data["name"]].set_data(data)
        else:
            self.row_data[data["name"]] = DeviceTableRow(data)
        # Update table cells
        with self.table_sort_on_hold:
            column_keys = list(self.headers_key_map.values())  # map columns
            for key, value in data.items():
                if key not in column_keys:
                    continue  # Skip userParameters and deviceConfig
                column = column_keys.index(key)
                item = self.table.item(row, column)
                if not item:
                    continue
                if key in ("enabled", "readOnly", "softwareTrigger"):
                    item.setCheckState(
                        QtCore.Qt.CheckState.Checked if value else QtCore.Qt.CheckState.Unchecked
                    )
                    item.setData(QtCore.Qt.ItemDataRole.UserRole, value)
                    item.setText("")  # No text for checkboxes
                elif key == "deviceTags":
                    item.setText(
                        ", ".join(value) if isinstance(value, (list, set, tuple)) else str(value)
                    )
                elif key == "deviceClass":
                    item.setText(
                        value.split(".")[-1]
                    )  # Only show the DeviceClass, not the full module
                else:
                    if value is None:
                        value = ""
                    item.setText(str(value))
            self._update_device_row_status(
                row,
                self.row_data[data["name"]].validation_status[0],
                self.row_data[data["name"]].validation_status[1],
            )
            self.table.resizeRowToContents(row)
        self._on_device_row_data_changed(self.row_data[data["name"]].data)
        return True

    def _update_device_row_status(
        self, row: int, config_status: int, connection_status: int
    ) -> bool:
        """
        Update an existing device row's validation status.

        Args:
            device_name (str): The name of the device.
            config_status (int): The configuration validation status.
            connection_status (int): The connection status.
        """
        with self.table_sort_on_hold:
            item = self.table.item(row, 0)  # Config status column
            if item:
                item.setData(QtCore.Qt.ItemDataRole.UserRole, config_status)
                item.setIcon(self._icons["config_status"][config_status])
            item = self.table.item(row, 1)  # Connect status column
            if item:
                item.setData(QtCore.Qt.ItemDataRole.UserRole, connection_status)
                item.setIcon(self._icons["connection_status"][connection_status])

            # Update the stored row data as well
            device_name = self._get_cell_data(row, 2)  # Name column
            device_row = self.row_data.get(device_name, None)
            if not device_row:
                return False
            device_row: DeviceTableRow
            device_row.set_validation_status(config_status, connection_status)
        return True

    def _get_cell_data(self, row: int, column: int) -> str | bool | None:
        """
        Get the data from a specific cell.

        Args:
            row (int): The row index.
            column (int): The column index.
        """
        item = self.table.item(row, column)
        if item is None:
            return None
        if column in (8, 9, 10):  # Checkboxes
            return item.checkState() == QtCore.Qt.CheckState.Checked
        return item.text()

    def _update_row(self, data: dict) -> int | None:
        """
        Update an existing row with new data.

        Args:
            data (dict): The device data to populate the row.
        Returns:
            int | None: The row index if updated, else None.
        """
        device_row = self.row_data.get(data.get("name"), {})
        if self._compare_configs(device_row.data, data):
            return None  # No update needed
        row = self._find_row_by_name(data.get("name", ""))
        if row is not None:
            self.__update_device_row_data(row, data)
        return row

    def _compare_configs(self, cfg1: dict, cfg2: dict) -> bool:
        """Compare two device configurations for equality."""
        try:
            cfg1_model = DeviceModel.model_validate(cfg1)
            cfg2_model = DeviceModel.model_validate(cfg2)
            return cfg1_model == cfg2_model
        except Exception as e:
            logger.error(f"Error comparing device configs: {e}")
            return False

    def _clear_table(self):
        """Remove all rows."""
        with self.table_sort_on_hold:
            n_rows = self.table.rowCount()
            for _ in range(n_rows):
                self.table.removeRow(0)
            self.row_data.clear()

    def _find_row_by_name(self, name: str) -> int | None:
        """
        Find a row by device name.

        Args:
            name (str): The name of the device to find.
        Returns:
            int | None: The row index if found, else None.
        """
        for row in range(self.table.rowCount()):
            data = self._get_cell_data(row, 2)
            if data and data == name:
                return row
        return None

    def _remove_rows_by_name(self, device_names: list[str]):
        """
        Remove a row by device name.

        Args:
            device_name (str): The name of the device to remove.
        """
        if not device_names:
            return
        with self.table_sort_on_hold:
            for device_name in device_names:
                row = self._find_row_by_name(device_name)
                if row is None:
                    logger.warning(f"Device {device_name} not found in table for removal.")
                    return
                self.table.removeRow(row)
                self.row_data.pop(device_name, None)

    def _is_config_in_sync_with_redis(self):
        """Check if the current config is in sync with Redis."""
        if (
            not self.client
            or not self.client.device_manager
            or not self.client.device_manager.devices
        ):
            return False  # No proper client connection
        redis_config = [
            DeviceModel.model_validate(device._config)
            for device in self.client.device_manager.devices.values()
        ]
        try:
            current_config = [
                DeviceModel.model_validate(row_data.data) for row_data in self.row_data.values()
            ]
            if redis_config == current_config:
                return True
            else:
                return False
        except Exception as e:
            logger.error(f"Error comparing device configs: {e}")
            return False

    def _get_overlapping_configs(self) -> list[dict[str, Any]]:
        """
        Get the device configs that overlap between the table and the config in the current running BEC session.
        A device will be ignored if it is disabled in the BEC session.

        Args:
            device_configs (Iterable[dict[str, Any]]): The device configs to check.

        Returns:
            list[dict[str, Any]]: The list of overlapping device configs.
        """
        overlapping_configs = []
        for cfg in self.get_device_config():
            device_name = cfg.get("name", None)
            if device_name is None:
                continue
            if self._is_device_in_redis_session(device_name, cfg):
                overlapping_configs.append(cfg)

        return overlapping_configs

    def _is_device_in_redis_session(self, device_name: str, device_config: dict) -> bool:
        """Check if a device is in the running section."""
        dev_obj = self.client.device_manager.devices.get(device_name, None)
        if dev_obj is None or dev_obj.enabled is False:
            return False
        return self._compare_device_configs(dev_obj._config, device_config)

    def _compare_device_configs(self, config1: dict, config2: dict) -> bool:
        """Compare two device configurations through the Device model in bec_lib.atlas_models.

        Args:
            config1 (dict): The first device configuration.
            config2 (dict): The second device configuration.

        Returns:
            bool: True if the configurations are equivalent, False otherwise.
        """
        try:
            model1 = DeviceModel.model_validate(config1)
            model2 = DeviceModel.model_validate(config2)
            return model1 == model2
        except Exception:
            return False

    # -------------------------------------------------------------------------
    # Public API to manage device configs in the table
    # -------------------------------------------------------------------------

    def get_device_config(self) -> list[dict]:
        """
        Get the current device configurations in the table.

        Returns:
            list[dict]: The list of device configurations.
        """
        cfgs = [
            {"name": device_name, **row_data.data}
            for device_name, row_data in self.row_data.items()
        ]
        return cfgs

    def get_validation_results(self) -> dict[str, Tuple[dict, int, int]]:
        """
        Get the current device validation results in the table.

        Returns:
            dict[str, Tuple[dict, int, int]]: Dictionary mapping of device name to
                                               (device config, config status, connection status).
        """
        return {
            row_data.data.get("name"): (row_data.data, *row_data.validation_status)
            for row_data in self.row_data.values()
            if row_data.data.get("name") is not None
        }

    def get_selected_device_configs(self) -> list[dict]:
        """
        Get the currently selected device configurations in the table.

        Returns:
            list[dict]: The list of selected device configurations.
        """
        selected_configs = []
        selected_rows = set()
        for index in self.table.selectionModel().selectedIndexes():
            selected_rows.add(index.row())
        for row in selected_rows:
            device_name = self._get_cell_data(row, 2)  # Name column
            if device_name:
                row_data = self.row_data.get(device_name)
                if row_data:
                    selected_configs.append(row_data.data)
        return selected_configs

    # -------------------------------------------------------------------------
    # Public API to be called via signals/slots
    # -------------------------------------------------------------------------

    @SafeSlot(list, bool)
    def set_device_config(self, device_configs: _DeviceCfgIter, skip_validation: bool = False):
        """
        Set the device config. This will clear any existing configs.

        Args:
            device_configs (Iterable[dict[str, Any]]): The device configs to set.
            skip_validation (bool): Whether to skip validation for the set devices.
        """
        self.set_busy(True)
        with self.table_sort_on_hold:
            self.clear_device_configs()
            cfgs_added = []
            for cfg in device_configs:
                self._add_row(cfg, ConfigStatus.UNKNOWN, ConnectionStatus.UNKNOWN)
                cfgs_added.append(cfg)
        self.device_configs_changed.emit(cfgs_added, True, skip_validation)
        in_sync_with_redis = self._is_config_in_sync_with_redis()
        self.device_config_in_sync_with_redis.emit(in_sync_with_redis)
        self.set_busy(False)

    @SafeSlot()
    def clear_device_configs(self):
        """Clear the device configs. Skips validation by default."""
        self.set_busy(True)
        device_configs = self.get_device_config()
        with self.table_sort_on_hold:
            self._clear_table()
        self.device_configs_changed.emit(
            device_configs, False, True
        )  # Skip validation for removals
        in_sync_with_redis = self._is_config_in_sync_with_redis()
        self.device_config_in_sync_with_redis.emit(in_sync_with_redis)
        self.set_busy(False)

    @SafeSlot(list, bool)
    def add_device_configs(self, device_configs: _DeviceCfgIter, skip_validation: bool = False):
        """
        Add devices to the config. If a device already exists, it will be replaced. If the validation is
        skipped, the device will be added with UNKNOWN state to the table and has to be manually adjusted
        by the user later on.

        Args:
            device_configs (Iterable[dict[str, Any]]): The device configs to add.
            skip_validation (bool): Whether to skip validation for the added devices.
        """
        self.set_busy(True)
        already_in_table = []
        not_in_table = []
        with self.table_sort_on_hold:
            for cfg in device_configs:
                if cfg["name"] in self.row_data:
                    already_in_table.append(cfg)
                else:
                    not_in_table.append(cfg)
            with self.table_sort_on_hold:
                # Remove existing rows first
                if len(already_in_table) > 0:
                    self._remove_rows_by_name([cfg["name"] for cfg in already_in_table])
                    self.device_configs_changed.emit(
                        already_in_table, False, True
                    )  # Skip validation for removals

                all_configs = already_in_table + not_in_table
                if len(all_configs) > 0:
                    for cfg in already_in_table + not_in_table:
                        self._add_row(cfg, ConfigStatus.UNKNOWN, ConnectionStatus.UNKNOWN)

        self.device_configs_changed.emit(already_in_table + not_in_table, True, skip_validation)
        in_sync_with_redis = self._is_config_in_sync_with_redis()
        self.device_config_in_sync_with_redis.emit(in_sync_with_redis)
        self.set_busy(False)

    @SafeSlot(list, bool)
    def update_device_configs(self, device_configs: _DeviceCfgIter, skip_validation: bool = False):
        """
        Update devices in the config. If a device does not exist, it will be added.

        Args:
            device_configs (Iterable[dict[str, Any]]): The device configs to update.
            skip_validation (bool): Whether to skip validation for the updated devices.
        """
        self.set_busy(True)
        cfgs_updated = []
        with self.table_sort_on_hold:
            for cfg in device_configs:
                if cfg["name"] not in self.row_data:
                    self._add_row(cfg, ConfigStatus.UNKNOWN, ConnectionStatus.UNKNOWN)
                    cfgs_updated.append(cfg)
                    continue
                # Update existing row if device config has changed
                row = self._update_row(cfg)
                if row is not None:
                    cfgs_updated.append(cfg)
        self.device_configs_changed.emit(cfgs_updated, True, skip_validation)
        in_sync_with_redis = self._is_config_in_sync_with_redis()
        self.device_config_in_sync_with_redis.emit(in_sync_with_redis)
        self.set_busy(False)

    @SafeSlot(list)
    def remove_device_configs(self, device_configs: _DeviceCfgIter):
        """
        Remove devices from the config.

        Args:
            device_configs (dict[str, dict]): The device configs to remove.
        """
        self.set_busy(True)
        cfgs_to_be_removed = list(device_configs)
        with self.table_sort_on_hold:
            self._remove_rows_by_name([cfg["name"] for cfg in cfgs_to_be_removed])
        self.device_configs_changed.emit(
            cfgs_to_be_removed, False, True
        )  # Skip validation for removals
        in_sync_with_redis = self._is_config_in_sync_with_redis()
        self.device_config_in_sync_with_redis.emit(in_sync_with_redis)
        self.set_busy(False)

    @SafeSlot(str)
    def remove_device(self, device_name: str):
        """
        Remove a device from the config.

        Args:
            device_name (str): The name of the device to remove.
        """
        self.set_busy(True)
        row_data = self.row_data.get(device_name)
        if not row_data:
            logger.warning(f"Device {device_name} not found in table for removal.")
            self.set_busy(False)
            return
        with self.table_sort_on_hold:
            self._remove_rows_by_name([row_data.data["name"]])
        cfgs = [{"name": device_name, **row_data.data}]
        self.device_configs_changed.emit(cfgs, False, True)  # Skip validation for removals
        in_sync_with_redis = self._is_config_in_sync_with_redis()
        self.device_config_in_sync_with_redis.emit(in_sync_with_redis)
        self.set_busy(False)

    @SafeSlot(list)
    def update_multiple_device_validations(self, validation_results: _ValidationResultIter):
        """
        Slot to update multiple device validation statuses. This is recommended and more
        efficient than updating individual device validation statuses which may affect
        the performance of the UI when many devices are being updated in quick succession.

        Args:
            device_configs (Iterable[dict[str, Any]]): The device configs to update.
        """
        self.set_busy(True)
        self.table.setSortingEnabled(False)
        logger.info(
            f"Updating multiple device validation statuses with names {[cfg.get('name', '') for cfg, _, _, _ in validation_results]}..."
        )
        for cfg, config_status, connection_status, _ in validation_results:
            logger.info(
                f"Updating device {cfg.get('name', '')} with config status {config_status} and connection status {connection_status}..."
            )
            row = self._find_row_by_name(cfg.get("name", ""))
            if row is None:
                logger.warning(f"Device {cfg.get('name')} not found in table for session update.")
                continue
            self._update_device_row_status(row, config_status, connection_status)
        in_sync_with_redis = self._is_config_in_sync_with_redis()
        self.device_config_in_sync_with_redis.emit(in_sync_with_redis)
        self.table.setSortingEnabled(True)
        self.set_busy(False)

    @SafeSlot(dict, int, int, str)
    def update_device_validation(
        self, device_config: dict, config_status: int, connection_status: int, validation_msg: str
    ) -> None:
        """
        Update the validation status of a device. If multiple devices are being updated in a batch,
        consider using the `update_multiple_device_validations` method instead for better performance.

        Args:

        """
        self.set_busy(True)
        row = self._find_row_by_name(device_config.get("name", ""))
        if row is None:
            logger.warning(
                f"Device {device_config.get('name')} not found in table for validation update."
            )
            self.set_busy(False)
            return
        # Disable here sorting without context manager to avoid triggering of registered
        # resizing methods. Those can be quite heavy, thus, should not run on every
        # update of a validation status.
        self.table.setSortingEnabled(False)
        self._update_device_row_status(row, config_status, connection_status)
        self.table.setSortingEnabled(True)
        in_sync_with_redis = self._is_config_in_sync_with_redis()
        self.device_config_in_sync_with_redis.emit(in_sync_with_redis)
        self.set_busy(False)



from bec_lib.logger import bec_logger
from qtpy.QtWidgets import QListWidget, QListWidgetItem, QWidget

logger = bec_logger.logger


class BECList(QListWidget):
    """List Widget that manages ListWidgetItems with associated widgets."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self._widget_map: dict[str, tuple[QListWidgetItem, QWidget]] = {}

    def __contains__(self, key: str) -> bool:
        return key in self._widget_map

    def add_widget_item(self, key: str, widget: QWidget):
        """
        Add a widget to the list, mapping is associated with the given key.

        Args:
            key (str): Key to associate with the widget.
            widget (QWidget): Widget to add to the list.
        """
        if key in self._widget_map:
            self.remove_widget_item(key)

        item = QListWidgetItem()
        item.setSizeHint(widget.sizeHint())
        self.insertItem(0, item)
        self.setItemWidget(item, widget)
        self._widget_map[key] = (item, widget)

    def remove_widget_item(self, key: str):
        """
        Remove a widget by identifier key.

        Args:
            key (str): Key associated with the widget to remove.
        """
        if key not in self._widget_map:
            return

        item, widget = self._widget_map.pop(key)
        row = self.row(item)
        self.takeItem(row)
        try:
            widget.close()
        except Exception:
            logger.debug(f"Could not close widget properly for key: {key}.")
        try:
            widget.deleteLater()
        except Exception:
            logger.debug(f"Could not delete widget properly for key: {key}.")

    def clear_widgets(self):
        """Remove and destroy all widget items."""
        for key in list(self._widget_map.keys()):
            self.remove_widget_item(key)
        self._widget_map.clear()
        self.clear()

    def get_widget(self, key: str) -> QWidget | None:
        """Return the widget for a given key."""
        entry = self._widget_map.get(key)
        return entry[1] if entry else None

    def get_item(self, key: str) -> QListWidgetItem | None:
        """Return the QListWidgetItem for a given key."""
        entry = self._widget_map.get(key)
        return entry[0] if entry else None

    def get_widgets(self) -> list[QWidget]:
        """Return all managed widgets."""
        return [w for _, w in self._widget_map.values()]

    def get_widget_for_item(self, item: QListWidgetItem) -> QWidget | None:
        """Return the widget associated with a given QListWidgetItem."""
        for itm, widget in self._widget_map.values():
            if itm == item:
                return widget
        return None

    def get_item_for_widget(self, widget: QWidget) -> QListWidgetItem | None:
        """Return the QListWidgetItem associated with a given widget."""
        for itm, w in self._widget_map.values():
            if w == widget:
                return itm
        return None

    def get_all_keys(self) -> list[str]:
        """Return all keys for managed widgets."""
        return list(self._widget_map.keys())



"""
Scan Control Dialog for Monaco Editor

This module provides a dialog wrapper around the ScanControl widget,
allowing users to configure and generate scan code that can be inserted
into the Monaco editor.
"""

from bec_lib.device import Device
from bec_lib.logger import bec_logger
from qtpy.QtCore import QSize, Qt
from qtpy.QtWidgets import QDialog, QDialogButtonBox, QPushButton, QVBoxLayout

from bec_widgets.widgets.control.scan_control import ScanControl

logger = bec_logger.logger


class ScanControlDialog(QDialog):
    """
    Dialog window containing the ScanControl widget for generating scan code.

    This dialog allows users to configure scan parameters and generates
    Python code that can be inserted into the Monaco editor.
    """

    def __init__(self, parent=None, client=None):
        super().__init__(parent)
        self.setWindowTitle("Insert Scan")

        # Store the client for passing to ScanControl
        self.client = client
        self._scan_code = ""

        self._setup_ui()

    def sizeHint(self) -> QSize:
        return QSize(600, 800)

    def _setup_ui(self):
        """Setup the dialog UI with ScanControl widget and buttons."""
        layout = QVBoxLayout(self)

        # Create the scan control widget
        self.scan_control = ScanControl(parent=self, client=self.client)
        self.scan_control.show_scan_control_buttons(False)
        layout.addWidget(self.scan_control)

        # Create dialog buttons
        button_box = QDialogButtonBox(Qt.Orientation.Horizontal, self)

        # Create custom buttons with appropriate text
        insert_button = QPushButton("Insert")
        cancel_button = QPushButton("Cancel")

        button_box.addButton(insert_button, QDialogButtonBox.ButtonRole.AcceptRole)
        button_box.addButton(cancel_button, QDialogButtonBox.ButtonRole.RejectRole)

        layout.addWidget(button_box)

        # Connect button signals
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)

    def _generate_scan_code(self):
        """Generate Python code for the configured scan."""
        try:
            # Get scan parameters from the scan control widget
            args, kwargs = self.scan_control.get_scan_parameters()
            scan_name = self.scan_control.current_scan

            if not scan_name:
                self._scan_code = ""
                return

            # Process arguments and add device prefix where needed
            processed_args = self._process_arguments_for_code_generation(args)
            processed_kwargs = self._process_kwargs_for_code_generation(kwargs)

            # Generate the Python code string
            code_parts = []

            # Process arguments and keyword arguments
            all_args = []

            # Add positional arguments
            if processed_args:
                all_args.extend(processed_args)

            # Add keyword arguments (excluding metadata)
            if processed_kwargs:
                kwargs_strs = [f"{k}={v}" for k, v in processed_kwargs.items()]
                all_args.extend(kwargs_strs)

            # Join all arguments and create the scan call
            args_str = ", ".join(all_args)
            if args_str:
                code_parts.append(f"scans.{scan_name}({args_str})")
            else:
                code_parts.append(f"scans.{scan_name}()")

            self._scan_code = "\n".join(code_parts)

        except Exception as e:
            logger.error(f"Error generating scan code: {e}")
            self._scan_code = f"# Error generating scan code: {e}\n"

    def _process_arguments_for_code_generation(self, args):
        """Process arguments to add device prefixes and proper formatting."""
        return [self._format_value_for_code(arg) for arg in args]

    def _process_kwargs_for_code_generation(self, kwargs):
        """Process keyword arguments to add device prefixes and proper formatting."""
        return {key: self._format_value_for_code(value) for key, value in kwargs.items()}

    def _format_value_for_code(self, value):
        """Format a single value for code generation."""
        if isinstance(value, Device):
            return f"dev.{value.name}"
        return repr(value)

    def get_scan_code(self) -> str:
        """
        Get the generated scan code.

        Returns:
            str: The Python code for the configured scan.
        """
        return self._scan_code

    def accept(self):
        """Override accept to generate code before closing."""
        self._generate_scan_code()
        super().accept()


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtpy.QtWidgets import QApplication

    app = QApplication(sys.argv)
    dialog = ScanControlDialog()
    dialog.show()
    sys.exit(app.exec_())



# Copyright (C) 2022 The Qt Company Ltd.
# SPDX-License-Identifier: LicenseRef-Qt-Commercial OR BSD-3-Clause

from qtpy.QtDesigner import QDesignerCustomWidgetInterface
from qtpy.QtWidgets import QWidget

from bec_widgets.utils.bec_designer import designer_material_icon
from bec_widgets.widgets.utility.toggle.toggle import ToggleSwitch

DOM_XML = """
<ui language='c++'>
    <widget class='ToggleSwitch' name='toggle_switch'>
    </widget>
</ui>
"""


class ToggleSwitchPlugin(QDesignerCustomWidgetInterface):  # pragma: no cover
    def __init__(self):
        super().__init__()
        self._form_editor = None

    def createWidget(self, parent):
        if parent is None:
            return QWidget()
        t = ToggleSwitch(parent)
        return t

    def domXml(self):
        return DOM_XML

    def group(self):
        return "BEC Utils"

    def icon(self):
        return designer_material_icon(ToggleSwitch.ICON_NAME)

    def includeFile(self):
        return "toggle_switch"

    def initialize(self, form_editor):
        self._form_editor = form_editor

    def isContainer(self):
        return False

    def isInitialized(self):
        return self._form_editor is not None

    def name(self):
        return "ToggleSwitch"

    def toolTip(self):
        return ""

    def whatsThis(self):
        return self.toolTip()
